# type: ignore
# -*- coding: utf-8 -*-
#
#       ramstk.analyses.milhdbk217f.models.Relay.py is part of the RAMSTK
#       Project
#
# All rights reserved.
# Copyright 2007 - 2017 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Relay MIL-HDBK-217F Calculations Module."""

# Standard Library Imports
from math import exp
from typing import Any, Dict

PART_COUNT_LAMBDA_B = {
    1: {
        1: [0.13, 0.28, 2.1, 1.1, 3.8, 1.1, 1.4, 1.9, 2.0, 7.0, 0.66, 3.5, 10.0, 0.0],
        2: [
            0.43,
            0.89,
            6.9,
            3.6,
            12.0,
            3.4,
            4.4,
            6.2,
            6.7,
            22.0,
            0.21,
            11.0,
            32.0,
            0.0,
        ],
        3: [0.13, 0.26, 2.1, 1.1, 3.8, 1.1, 1.4, 1.9, 2.0, 7.0, 0.66, 3.5, 10.0, 0.0],
        4: [0.11, 0.23, 1.8, 0.92, 3.3, 0.96, 1.2, 2.1, 2.3, 6.5, 0.54, 3.0, 9.0, 0.0],
        5: [0.29, 0.60, 4.8, 2.4, 8.2, 2.3, 2.9, 4.1, 4.5, 15.0, 0.14, 7.6, 22.0, 0.0],
        6: [
            0.88,
            1.8,
            14.0,
            7.4,
            26.0,
            7.1,
            9.1,
            13.0,
            14.0,
            46.0,
            0.44,
            24.0,
            67.0,
            0.0,
        ],
    },
    2: {
        1: [0.40, 1.2, 4.8, 2.4, 6.8, 4.8, 7.6, 8.4, 13.0, 9.2, 0.16, 4.8, 13.0, 240.0],
        2: [
            0.50,
            1.5,
            6.0,
            3.0,
            8.5,
            5.0,
            9.5,
            11.0,
            16.0,
            12.0,
            0.20,
            5.0,
            17.0,
            300.0,
        ],
    },
}
PART_COUNT_PI_Q = {1: [0.6, 3.0, 9.0], 2: [0.0, 1.0, 4.0]}
PART_STRESS_PI_Q = {1: [0.1, 0.3, 0.45, 0.6, 1.0, 1.5, 3.0], 2: [1.0, 4.0]}
PI_C = {1: [1.0, 1.5, 1.75, 2.0, 2.5, 3.0, 4.25, 5.5, 8.0]}
PI_E = {
    1: {
        1: [
            1.0,
            2.0,
            15.0,
            8.0,
            27.0,
            7.0,
            9.0,
            11.0,
            12.0,
            46.0,
            0.50,
            25.0,
            66.0,
            0.0,
        ],
        2: [
            2.0,
            5.0,
            44.0,
            24.0,
            78.0,
            15.0,
            20.0,
            28.0,
            38.0,
            140.0,
            1.0,
            72.0,
            200.0,
            0.0,
        ],
    },
    2: [
        1.0,
        3.0,
        12.0,
        6.0,
        17.0,
        12.0,
        19.0,
        21.0,
        32.0,
        23.0,
        0.4,
        12.0,
        33.0,
        590.0,
    ],
}
PI_F = {
    1: {
        1: {
            1: [4.0, 8.0],
            2: [6.0, 18.0],
            3: [1.0, 3.0],
            4: [4.0, 8.0],
            5: [7.0, 14.0],
            6: [7.0, 14.0],
        }
    },
    2: {
        1: {1: [3.0, 6.0], 2: [5.0, 10.0], 3: [6.0, 12.0]},
        2: {
            1: [5.0, 10.0],
            2: [2.0, 6.0],
            3: [6.0, 12.0],
            4: [100.0, 100.0],
            5: [10.0, 20.0],
        },
        3: {1: [10.0, 20.0], 2: [100.0, 100.0]},
        4: {1: [6.0, 12.0], 2: [1.0, 3.0]},
        5: {1: [25.0, 0.0], 2: [6.0, 0.0]},
        6: {1: [10.0, 20.0]},
        7: {1: [9.0, 12.0]},
        8: {1: [10.0, 20.0], 2: [5.0, 10.0], 3: [5.0, 10.0]},
    },
    3: {
        1: {1: [20.0, 40.0], 2: [5.0, 10.0]},
        2: {
            1: [3.0, 6.0],
            2: [1.0, 3.0],
            3: [2.0, 6.0],
            4: [3.0, 6.0],
            5: [2.0, 6.0],
            6: [2.0, 6.0],
        },
    },
    4: {1: {1: [7.0, 14.0], 2: [12.0, 24.0], 3: [10.0, 20.0], 4: [5.0, 10.0]}},
}


def calculate_cycling_factor(quality_id: int, n_cycles: float) -> float:
    """Calculate the cycling factor (piCYC) for the relay.

    :param quality_id: the quality level identifier.
    :param n_cycles: the number of relay cycles per hour in application.
    :return: _pi_cyc; the calculate cycling factor.
    :rtype: float
    """
    if quality_id in [1, 2, 3, 4, 5, 6] and n_cycles < 1.0:
        _pi_cyc = 0.1
    elif quality_id == 7 and n_cycles > 1000.0:
        _pi_cyc = (n_cycles / 100.0) ** 2.0
    elif quality_id == 7 and 10.0 < n_cycles < 1000.0:
        _pi_cyc = n_cycles / 10.0
    else:
        _pi_cyc = 0.0

    return _pi_cyc


def calculate_load_stress_factor(technology_id: int, current_ratio: float) -> float:
    """Calculate the load stress factor (piL).

    Only subcategory 1 relays use this in their calculation.

    :param technology_id: the relay technology identifier.
    :param current_ratio: the operating current ratio of the relay.
    :return: _pi_l; the calculate value of piL.
    :rtype: float
    """
    if technology_id == 1:
        _pi_l = (current_ratio / 0.8) ** 2.0
    elif technology_id == 2:
        _pi_l = (current_ratio / 0.4) ** 2.0
    elif technology_id == 3:
        _pi_l = (current_ratio / 0.2) ** 2.0
    else:
        _pi_l = 0.0

    return _pi_l


def calculate_part_count(**attributes: Dict[str, Any]) -> float:
    """Wrap get_part_count_lambda_b().

    This wrapper allows us to pass an attributes dict from a generic parts
    count function.

    :param attributes: the attributes for the connection being calculated.
    :return: _base_hr; the parts count base hazard rates.
    :rtype: float
    """
    return get_part_count_lambda_b(
        subcategory_id=attributes["subcategory_id"],
        type_id=attributes["type_id"],
        environment_active_id=attributes["environment_active_id"],
    )


def calculate_part_stress(**attributes: Dict[str, Any]) -> Dict[str, Any]:
    """Calculate the part stress hazard rate for a relay.

    This function calculates the MIL-HDBK-217F hazard rate using the part
    stress method.

    :return: (attributes, _msg); the keyword argument (hardware attribute)
             dictionary with updated values and the error message, if any.
    :rtype: (dict, str)
    """
    attributes["lambda_b"] = calculate_part_stress_lambda_b(
        attributes["subcategory_id"],
        attributes["type_id"],
        attributes["temperature_active"],
    )
    attributes["piCYC"] = calculate_cycling_factor(
        attributes["quality_id"], attributes["n_cycles"]
    )
    attributes["piL"] = calculate_load_stress_factor(
        attributes["technology_id"], attributes["current_ratio"]
    )
    attributes["piE"] = get_environment_factor(
        attributes["subcategory_id"],
        attributes["quality_id"],
        attributes["environment_active_id"],
    )
    attributes["piF"] = get_application_construction_factor(
        attributes["quality_id"],
        attributes["contact_rating_id"],
        attributes["construction_id"],
        attributes["application_id"],
    )

    if attributes["subcategory_id"] == 1:
        attributes["piC"] = PI_C[attributes["subcategory_id"]][
            attributes["contact_form_id"] - 1
        ]

    attributes["hazard_rate_active"] = (
        attributes["lambda_b"] * attributes["piQ"] * attributes["piE"]
    )
    if attributes["subcategory_id"] == 1:
        attributes["hazard_rate_active"] = (
            attributes["hazard_rate_active"]
            * attributes["piL"]
            * attributes["piC"]
            * attributes["piCYC"]
            * attributes["piF"]
        )

    return attributes


def calculate_part_stress_lambda_b(
    subcategory_id: int, type_id: int, temperature_active: float
) -> float:
    """Calculate the base hazard rate for the relay.

    :param subcategory_id: the subcategory identifier.
    :param type_id: the relay type identifier.
    :param temperature_active: the operating ambient temperature of the
        relay in C.
    :return: _lambda_b; the calculated part stress base hazard rate or 0.0 if
        passed an unknown subcategory ID.
    :rtype: float
    """
    _dic_factors = {
        1: [[0.00555, 352.0, 15.7], [0.0054, 377.0, 10.4]],
        2: [0.4, 0.5, 0.5],
    }
    _lambda_b = 0.0

    if subcategory_id == 1:
        _f0 = _dic_factors[subcategory_id][type_id - 1][0]
        _f1 = _dic_factors[subcategory_id][type_id - 1][1]
        _f2 = _dic_factors[subcategory_id][type_id - 1][2]
        _lambda_b = _f0 * exp(((temperature_active + 273.0) / _f1) ** _f2)
    elif subcategory_id == 2:
        _lambda_b = _dic_factors[subcategory_id][type_id - 1]

    return _lambda_b


def get_application_construction_factor(
    quality_id: int, contact_rating_id: int, construction_id: int, application_id: int
) -> float:
    """Calculate the construction factor (piF) for the relay.

    :param quality_id: the quality level identifier.
    :param contact_rating_id: the relay contact current rating identifier.
    :param construction_id: the relay construction identifier.
    :param application_id: the relay application identifier.
    :return: _pi_f; the selected application factor.
    :rtype: float
    :raise: KeyError if passed an unknown contact rating ID, construction ID,
        or application ID.
    """
    if quality_id in [1, 2, 3, 4, 5, 6]:
        _quality = 1
    else:
        _quality = 2

    return PI_F[contact_rating_id][application_id][construction_id][_quality - 1]


def get_environment_factor(
    subcategory_id: int, quality_id: int, environment_active_id: int
) -> float:
    """Retrieve the environment factor (pi_E).

    :param subcategory_id: the subcategory identifier.
    :param quality_id: the quality level identifier.
    :param environment_active_id: the active environment identifier.
    :return: _pi_e; the selected environment factor.
    :rtype: float
    :raise: IndexError if passed an unknown active environment ID.
    :raise: KeyError if passed an unknown subcategory ID.
    """
    if quality_id in [1, 2, 3, 4, 5, 6]:
        _quality = 1
    else:
        _quality = 2

    if subcategory_id == 1:
        _pi_e = PI_E[subcategory_id][_quality][environment_active_id - 1]
    else:
        _pi_e = PI_E[subcategory_id][environment_active_id - 1]

    return _pi_e


def get_part_count_lambda_b(**kwargs: Dict[str, int]) -> float:
    """Retrieve the parts count base hazard rate (lambda b) from MIL-HDBK-217F.

    This function calculates the MIL-HDBK-217F hazard rate using the parts
    count method.

    This function calculates the MIL-HDBK-217F hazard rate using the parts
    count method.  The dictionary PART_COUNT_217F_LAMBDA_B contains the
    MIL-HDBK-217F parts count base hazard rates.  Keys are for
    PART_COUNT_217F_LAMBDA_B are:

        #. subcategory_id
        #. environment_active_id
        #. type id; if the relay subcategory is NOT type dependent, then
            the second key will be zero.

    Current subcategory IDs are:

    +----------------+-------------------------------+-----------------+
    | Subcategory    |              Relay            | MIL-HDBK-217F   |
    |       ID       |              Style            |    Section      |
    +================+===============================+=================+
    |        1       | Mechanical                    |       13.1      |
    +----------------+-------------------------------+-----------------+
    |        2       | Solid State                   |       13.2      |
    +----------------+-------------------------------+-----------------+

    :param subcategory_id: the subcategory identifier.
    :param type_id: the relay type identifier.
    :param environment_active_id: the active environment identifier.
    :return: _base_hr; the parts count base hazard rate.
    :rtype: float
    :raise: IndexError if passed an unknown active environment ID.
    :raise: KeyError if passed an unknown subcategory ID or type ID.
    """
    _subcategory_id = kwargs.get("subcategory_id", 0)
    _type_id = kwargs.get("type_id", 0)
    _environment_active_id = kwargs.get("environment_active_id", 0)

    return PART_COUNT_LAMBDA_B[_subcategory_id][_type_id][_environment_active_id - 1]



# pylint: disable=unused-import
# -*- coding: utf-8 -*-
#
#       ramstk.__init__.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright 2007 - 2020 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Reliability, Availability, Maintainability, and Safety ToolKit package."""

# RAMSTK Local Imports
from .configuration import (
    RAMSTK_ACTIVE_ENVIRONMENTS,
    RAMSTK_ALLOCATION_MODELS,
    RAMSTK_CONTROL_TYPES,
    RAMSTK_COST_TYPES,
    RAMSTK_CRITICALITY,
    RAMSTK_DORMANT_ENVIRONMENTS,
    RAMSTK_FAILURE_PROBABILITY,
    RAMSTK_HR_DISTRIBUTIONS,
    RAMSTK_HR_MODELS,
    RAMSTK_HR_TYPES,
    RAMSTK_LIFECYCLE,
    RAMSTK_MTTR_TYPES,
    RAMSTK_S_DIST,
    RAMSTK_SW_APPLICATION,
    RAMSTK_SW_DEV_ENVIRONMENTS,
    RAMSTK_SW_DEV_PHASES,
    RAMSTK_SW_LEVELS,
    RAMSTK_SW_TEST_METHODS,
    RAMSTKSiteConfiguration,
    RAMSTKUserConfiguration,
)
from .logger import RAMSTKLogManager
from .utilities import (
    boolean_to_integer,
    date_to_ordinal,
    deprecated,
    dir_exists,
    file_exists,
    get_install_prefix,
    integer_to_boolean,
    none_to_default,
    none_to_string,
    ordinal_to_date,
    split_string,
    string_to_boolean,
)



# Standard Library Imports
from typing import Any

# Third Party Imports
from sqlalchemy.orm import relationship

# RAMSTK Package Imports
from ramstk.db import RAMSTK_BASE as RAMSTK_BASE
from ramstk.models import RAMSTKBaseRecord as RAMSTKBaseRecord

class RAMSTKCauseRecord(RAMSTK_BASE, RAMSTKBaseRecord):
    __defaults__: Any
    __tablename__: str
    __table_args__: Any
    revision_id: Any
    hardware_id: Any
    mode_id: Any
    mechanism_id: Any
    cause_id: Any
    description: Any
    rpn: Any
    rpn_detection: Any
    rpn_detection_new: Any
    rpn_new: Any
    rpn_occurrence: Any
    rpn_occurrence_new: Any
    mechanism: relationship
    control: relationship
    action: relationship
    is_mode: bool
    is_mechanism: bool
    is_cause: bool
    is_control: bool
    is_action: bool
    def get_attributes(self): ...



# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.fmea.view.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""GTK3 FMEA Views."""

# Standard Library Imports
from typing import Any, Dict, List

# Third Party Imports
from pubsub import pub

# RAMSTK Package Imports
from ramstk.configuration import (
    RAMSTK_CONTROL_TYPES,
    RAMSTK_CRITICALITY,
    RAMSTK_FAILURE_PROBABILITY,
    RAMSTKUserConfiguration,
)
from ramstk.logger import RAMSTKLogManager
from ramstk.views.gtk3 import Gtk, _
from ramstk.views.gtk3.assistants import AddControlAction
from ramstk.views.gtk3.widgets import RAMSTKMessageDialog, RAMSTKPanel, RAMSTKWorkView

# RAMSTK Local Imports
from . import FMEAMethodPanel, FMEATreePanel


def do_request_insert(level: str, parent_id: str) -> None:
    """Send the correct request for the FMEA item to insert.

    :param level: the indenture level in the FMEA of the new element to
        insert.
    :param parent_id: the node ID in the FMEA treelib Tree() of the
        parent element.
    :return: None
    :rtype: None
    """
    if level == "mode":
        pub.sendMessage("request_insert_fmea_mode")
    elif level == "mechanism":
        pub.sendMessage("request_insert_fmea_mechanism", mode_id=str(parent_id))
    elif level == "cause":
        pub.sendMessage("request_insert_fmea_cause", parent_id=str(parent_id))
    elif level in ["control", "action"]:
        pub.sendMessage(f"request_insert_fmea_{level}", parent_id=str(parent_id))


class FMEAWorkView(RAMSTKWorkView):
    """Display FMEA attribute data in the Work Book.

    The WorkView displays all the attributes for the Failure Mode and Effects
    Analysis (FMEA). The attributes of a FMEA Work View are:

    :cvar dict _dic_column_masks: dict with the list of masking values for
        the FMEA worksheet.  Key is the FMEA indenture level, value is a
        list of True/False values for each column in the worksheet.
    :cvar dict _dic_headings: dict with the variable headings for the first two
        columns.  Key is the name of the FMEA indenture level, value is a list
        of heading text.
    :cvar dict _dic_keys:
    :cvar dict _dic_column_keys:
    :cvar list _lst_control_type: list containing the types of controls that
        can be implemented.
    :cvar list _lst_labels: list containing the label text for each widget
        label.
    :cvar bool _pixbuf: indicates whether or icons are displayed in the
        RAMSTKTreeView.  If true, a GDKPixbuf column will be appended when
        creating the RAMSTKTreeView.  Default is True.
    :cvar str _tag: the name of the module.

    :ivar list _lst_callbacks: the list of callback methods for the view's
        toolbar buttons and pop-up menu.  The methods are listed in the order
        they appear on the toolbar and pop-up menu.
    :ivar list _lst_icons: the list of icons for the view's toolbar buttons
        and pop-up menu.  The icons are listed in the order they appear on the
        toolbar and pop-up menu.
    :ivar list _lst_mnu_labels: the list of labels for the view's pop-up
        menu.  The labels are listed in the order they appear in the menu.
    :ivar list _lst_tooltips: the list of tooltips for the view's
        toolbar buttons and pop-up menu.  The tooltips are listed in the
        order they appear on the toolbar or pop-up menu.
    :ivar dict _dic_missions: dict containing all this missions associated
        with the selected Revision.
    :ivar dict _dic_mission_phases: dict containing all the mission phases
        associated with each mission in _dic_missions.
    :ivar float _item_hazard_rate: hazard rate of the Hardware item associated
        with the FMEA.
    """

    # Define private dictionary attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _tag: str = "fmea"
    _pixbuf: bool = True
    _tablabel: str = _("FMEA")
    _tabtooltip: str = _(
        "Displays failure mode and effects analysis (FMEA) information for the "
        "selected Hardware item."
    )

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(
        self, configuration: RAMSTKUserConfiguration, logger: RAMSTKLogManager
    ) -> None:
        """Initialize the Work View for the FMEA.

        :param configuration: the RAMSTKUserConfiguration class instance.
        :param logger: the RAMSTKLogManager class instance.
        """
        super().__init__(configuration, logger)

        # Initialize private dictionary attributes.

        # Initialize private list attributes.
        self._lst_callbacks.insert(0, self._do_request_insert_sibling)
        self._lst_callbacks.insert(1, self._do_request_insert_child)
        self._lst_callbacks.insert(2, self._do_request_delete)
        self._lst_callbacks.insert(3, self._do_request_calculate)
        self._lst_icons.insert(0, "insert_sibling")
        self._lst_icons.insert(1, "insert_child")
        self._lst_icons.insert(2, "remove")
        self._lst_icons.insert(3, "calculate")
        self._lst_mnu_labels.insert(0, _("Add Sibling"))
        self._lst_mnu_labels.insert(1, _("Add Child"))
        self._lst_mnu_labels.insert(2, _("Delete Selected"))
        self._lst_mnu_labels.insert(3, _("Calculate FMEA"))
        self._lst_tooltips: List[str] = [
            _(
                "Add a new (D)FME(C)A entity at the same level as the "
                "currently selected entity."
            ),
            _(
                "Add a new (D)FME(C)A entity one level below the currently "
                "selected entity."
            ),
            _("Delete the selected entity from the (D)FME(C)A."),
            _(
                "Calculate the Task 102 criticality and/or risk priority "
                "number (RPN)."
            ),
            _("Save changes to the selected entity in the (D)FME(C)A."),
            _("Save changes to all entities in the (D)FME(C)A."),
        ]

        # Initialize private scalar attributes.
        self._item_hazard_rate: float = 0.0

        self._pnlMethods: RAMSTKPanel = FMEAMethodPanel()
        self._pnlPanel: RAMSTKPanel = FMEATreePanel()

        # Initialize public dictionary attributes.

        # Initialize public list attributes.

        # Initialize public scalar attributes.

        self.__make_ui()

        # Subscribe to PyPubSub messages.
        pub.subscribe(self._do_set_record_id, "selected_fmea")
        pub.subscribe(
            self._on_get_hardware_attributes, "succeed_get_all_hardware_attributes"
        )

    def _do_request_calculate(self, __button: Gtk.ToolButton) -> None:
        """Calculate the FMEA RPN or criticality.

        :param __button: the Gtk.ToolButton() that called this method.
        :return: None
        :rtype: None
        """
        if self._pnlMethods.chkCriticality.get_active():
            pub.sendMessage(
                "request_calculate_criticality", item_hr=self._item_hazard_rate
            )

        if self._pnlMethods.chkRPN.get_active():
            pub.sendMessage("request_calculate_rpn", method="mechanism")

    def _do_request_delete(self, __button: Gtk.ToolButton) -> None:
        """Request to delete the selected entity from the FMEA.

        :param __button: the Gtk.ToolButton() that called this method.
        :return: None
        :rtype: None
        """
        _parent = self.get_parent().get_parent().get_parent().get_parent()
        _model, _row = self._pnlPanel.tvwTreeView.get_selection().get_selected()
        _node_id = _model.get_value(_row, 0)

        _prompt = _(
            "You are about to delete {1} item {0} and all "
            "data associated with it.  Is this really what "
            "you want to do?"
        ).format(_node_id, self._tag.title())
        _dialog = RAMSTKMessageDialog(parent=_parent)
        _dialog.do_set_message(_prompt)
        _dialog.do_set_message_type("question")

        if _dialog.do_run() == Gtk.ResponseType.YES:
            super().do_set_cursor_busy()
            pub.sendMessage(
                "request_delete_fmea",
                node_id=_node_id,
            )

        _dialog.do_destroy()

    def _do_request_insert_child(self, __button: Gtk.ToolButton) -> None:
        """Request to insert a new entity to the FMEA.

        :return: None
        :rtype: None
        """
        # Try to get the information needed to add a new entity at the correct
        # location in the FMEA.  If there is nothing in the FMEA, by default
        # add a failure Mode.
        _model, _row = self._pnlPanel.tvwTreeView.get_selection().get_selected()
        try:
            _parent_id = _model.get_value(_row, 0)
            _level = {
                1: "mechanism",
                2: "cause",
                3: "control_action",
            }[len(str(_parent_id).split("."))]
        except TypeError:
            _parent_id = "0"
            _level = "mechanism"

        if _level == "control_action":
            _level = self.__on_request_insert_control_action()

        super().do_set_cursor_busy()

        # noinspection PyArgumentList
        do_request_insert(_level, _parent_id)

    def _do_request_insert_sibling(self, __button: Gtk.ToolButton) -> None:
        """Request to insert a new entity to the FMEA.

        :return: None
        :rtype: None
        """
        # Try to get the information needed to add a new entity at the correct
        # location in the FMEA.  If there is nothing in the FMEA, by default
        # add a failure Mode.
        _model, _row = self._pnlPanel.tvwTreeView.get_selection().get_selected()
        try:
            _parent_id = _model.get_value(_model.iter_parent(_row), 0)
            _level = {
                0: "mode",
                1: "mechanism",
                2: "cause",
                3: "control_action",
            }[len(str(_parent_id).split("."))]
        except TypeError:
            _parent_id = "0"
            _level = "mode"

        if _level == "control_action":
            _level = self.__on_request_insert_control_action()

        super().do_set_cursor_busy()
        # noinspection PyArgumentList
        do_request_insert(_level, _parent_id)

    def _do_set_record_id(self, attributes: Dict[str, Any]) -> None:
        """Set the record and revision ID when a hardware item is selected.

        :param attributes: the hazard dict for the selected hardware ID.
        :return: None
        :rtype: None
        """
        self.dic_pkeys["record_id"] = attributes["node_id"]

    def _on_get_hardware_attributes(self, attributes: Dict[str, Any]) -> None:
        """Set the hardware item hazard rate.

        :param attributes:
        :return:
        """
        self._item_hazard_rate = attributes["hazard_rate_active"]

    def __do_load_action_lists(self):
        """Load the Gtk.CellRendererCombo()s associated with FMEA actions.

        :return: None
        :rtype: None
        """
        self._pnlPanel.lst_action_category = [
            x[1][1]
            for x in self.RAMSTK_USER_CONFIGURATION.RAMSTK_ACTION_CATEGORY.items()
        ]
        self._pnlPanel.lst_action_status = [
            x[1][0] for x in self.RAMSTK_USER_CONFIGURATION.RAMSTK_ACTION_STATUS.items()
        ]
        self._pnlPanel.lst_users = [
            x[1][0] + ", " + x[1][1]
            for x in self.RAMSTK_USER_CONFIGURATION.RAMSTK_USERS.items()
        ]
        self._pnlPanel.lst_control_types = RAMSTK_CONTROL_TYPES

    def __do_load_rpn_lists(self):
        """Load the Gtk.CellRendererCombo()s associated with RPNs.

        :return: None
        :rtype: None
        """
        self._pnlPanel.lst_rpn_detection = [
            x[1]["name"]
            for x in self.RAMSTK_USER_CONFIGURATION.RAMSTK_RPN_DETECTION.items()
        ]
        self._pnlPanel.lst_rpn_detection.insert(0, "")
        self._pnlPanel.lst_rpn_occurrence = [
            x[1]["name"]
            for x in self.RAMSTK_USER_CONFIGURATION.RAMSTK_RPN_OCCURRENCE.items()
        ]
        self._pnlPanel.lst_rpn_occurrence.insert(0, "")
        self._pnlPanel.lst_rpn_severity = [
            x[1]["name"]
            for x in self.RAMSTK_USER_CONFIGURATION.RAMSTK_RPN_SEVERITY.items()
        ]
        self._pnlPanel.lst_rpn_severity.insert(0, "")

    def __do_load_severity_lists(self):
        """Load the Gtk.CellRendererCombo()s associated with CA risk.

        :return: None
        :rtype: None
        """
        self._pnlPanel.lst_mode_probability = [x[0] for x in RAMSTK_FAILURE_PROBABILITY]
        self._pnlPanel.lst_severity_class = [x[0] for x in RAMSTK_CRITICALITY]

    def __make_ui(self) -> None:
        """Build the user interface for the FMEA tab.

        :return: None
        :rtype: None
        """
        _hpaned: Gtk.HPaned = super().do_make_layout_lr()

        self._pnlPanel.dic_icons = self._dic_icons

        super().do_embed_treeview_panel()
        self.__do_load_action_lists()
        self.__do_load_rpn_lists()
        self.__do_load_severity_lists()
        self._pnlPanel.do_load_comboboxes()

        self.remove(self.get_children()[-1])
        _hpaned.pack1(self._pnlMethods, True, True)
        _hpaned.pack2(self._pnlPanel, True, True)

        self.show_all()

    def __on_request_insert_control_action(self) -> str:
        """Raise dialog to select whether to add a control or action.

        :return: _level; the level to add, control or action.
        :rtype: str
        """
        _level = ""

        _dialog = AddControlAction(
            parent=self.get_parent().get_parent().get_parent().get_parent()
        )

        if _dialog.do_run() == Gtk.ResponseType.OK:
            if _dialog.rdoControl.get_active():
                _level = "control"
            elif _dialog.rdoAction.get_active():
                _level = "action"

        _dialog.do_destroy()

        return _level



# -*- coding: utf-8 -*-
#
#       ramstk.models.commondb.RAMSTKStakeholders.py is part of The RAMSTK
#       Project
#
# All rights reserved.
# Copyright 2007 - 2017 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""RAMSTKStakeholders Table Module."""

# Third Party Imports
from sqlalchemy import Column, Integer, String

# RAMSTK Package Imports
from ramstk.db import RAMSTK_BASE
from ramstk.models import RAMSTKBaseRecord


class RAMSTKStakeholders(RAMSTK_BASE, RAMSTKBaseRecord):
    """Class to represent ramstk_stakeholders in the RAMSTK Common database."""

    __defaults__ = {"stakeholder": "Stakeholder"}
    __tablename__ = "ramstk_stakeholders"
    __table_args__ = {"extend_existing": True}

    stakeholders_id = Column(
        "fld_stakeholders_id",
        Integer,
        primary_key=True,
        autoincrement=True,
        nullable=False,
    )
    stakeholder = Column(
        "fld_stakeholder", String(512), default=__defaults__["stakeholder"]
    )

    def get_attributes(self):
        """Retrieve current values of RAMSTKStakeholders data model attributes.

        :return: {stakeholders_id, stakeholder} pairs.
        :rtype: dict
        """
        _attributes = {
            "stakeholders_id": self.stakeholders_id,
            "stakeholder": self.stakeholder,
        }

        return _attributes



# Standard Library Imports
from typing import Any, Dict

# RAMSTK Package Imports
from ramstk.models import RAMSTKBaseView

class RAMSTKPoFView(RAMSTKBaseView):
    _root: int = ...
    _tag: str = ...
    _dic_load_functions: Any = ...
    _dic_trees: Any = ...
    _lst_modules: Any = ...
    def __init__(self, **kwargs: Dict[Any, Any]) -> None: ...
    def _do_load_mechanisms(self) -> None: ...
    def _do_load_oploads(self, mechanism_id: int) -> None: ...
    def _do_load_opstress(self, load_id: int, parent_id: str) -> None: ...
    def _do_load_test_method(self, load_id: int, parent_id: str) -> None: ...



# type: ignore
# -*- coding: utf-8 -*-
#
#       ramstk.analyses.statistics.exponential.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Exponential Module."""

# Standard Library Imports
from typing import Tuple

# Third Party Imports
import scipy
from scipy.stats import expon


def get_hazard_rate(scale: float, location: float = 0.0) -> float:
    """Calculates the hazard rate given a scale and location parameter.

    This function calculates the rate parameter (lambda) given the scale (MTBF)
    and, optionally, a location parameter.

        >>> get_hazard_rate(10000.0)
        0.0001

        >>> get_hazard_rate(0.0)
        nan

    :param scale: the scale (MTBF) parameter.
    :param location: the location parameter.
    :return: _hazard_rate; the hazard rate.
    :rtype: float
    """
    _hazard_rate = 1.0 / expon.mean(
        loc=location,
        scale=scale,
    )

    return _hazard_rate


def get_mtbf(rate: float, location: float = 0.0) -> float:
    """Calculate the MTBF (scale) given a hazard rate (lambda) and location parameter.

        >>> get_mtbf(0.0008691)
        1150.6155793349442

        >>> get_mtbf(0.0)
        0.0

    :param rate: the rate (lambda) parameter.
    :param location: the location parameter.
    :return: _mtbf; the MTBF.
    :rtype: float
    """
    try:
        _mtbf = expon.mean(
            loc=location,
            scale=1.0 / rate,
        )
    except ZeroDivisionError:
        _mtbf = 0.0

    return _mtbf


def get_survival(scale: float, time: float, location: float = 0.0) -> float:
    """Calculate value of the survival function at time given scale and location.

    This function returns the lower and upper alpha bounds as well as the point
    estimate of the survival function at time.

        >>> get_survival(10000.0, 4.0)
        0.9996000799893344

        >>> get_survival(10000.0, 4.0, location=1.0)
        0.9997000449955004

    :param scale: the point estimate of the scale parameter.
    :param time: the time at which to calculate the survival function.
    :param location: the point estimate of the location parameter.
    :return: _surv; the value of the survival function at time.
    :rtype: float
    """
    return expon.sf(time, loc=location, scale=scale)


def do_fit(data, **kwargs) -> Tuple[float, float]:
    """Fits the provided data to the EXP distribution and estimates scale and location.

    :param data: the data to use in the fit.
    :return: (_location, _scale); the estimated parameters.
    :rtype: tuple
    """
    _location = kwargs.get("location", 0.0)  # Initial guess for location.
    _floc = kwargs.get("floc", None)  # Value to fix location parameter.
    _scale = kwargs.get("scale", 0.0)  # Initial guess for scale.
    _method = kwargs.get("method", "MLE")  # One of MLE or MM.

    # method is not an argument to fit() until scipy-1.7.1.
    if _floc is None:
        if scipy.__version__ >= "1.7.1":
            _location, _scale = expon.fit(
                data,
                loc=_location,
                scale=_scale,
                method=_method,
            )
        else:
            _location, _scale = expon.fit(
                data,
                loc=_location,
                scale=_scale,
            )
    else:
        if scipy.__version__ >= "1.7.1":
            _location, _scale = expon.fit(
                data,
                loc=_location,
                floc=_floc,
                scale=_scale,
                method=_method,
            )
        else:
            _location, _scale = expon.fit(
                data,
                loc=_location,
                floc=_floc,
                scale=_scale,
            )

    return _location, _scale



# -*- coding: utf-8 -*-
#
#       ramstk.analyses.improvementfactor.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright 2019 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Stakeholder Improvement Factor Analysis Module."""

# Standard Library Imports
from typing import Dict, Tuple


def calculate_improvement(
    planned_rank: int, customer_rank: int, priority: int, **kwargs: Dict[str, float]
) -> Tuple[float, float]:
    """Calculate the stakeholder improvement factor and overall weighting.

    :param planned_rank: the rank the design team plans.
    :param customer_rank: the rank the customer places on the input.
    :param priority: the design team's priority.
    :return: (_improvement, _weight); the improvement factor and overall
        weighting of the stakeholder input.
    :rtype: tuple
    """
    _user_float_1 = kwargs.get("user_float_1", 1.0)
    _user_float_2 = kwargs.get("user_float_2", 1.0)
    _user_float_3 = kwargs.get("user_float_3", 1.0)
    _user_float_4 = kwargs.get("user_float_4", 1.0)
    _user_float_5 = kwargs.get("user_float_5", 1.0)

    _improvement = 1.0 + 0.2 * (planned_rank - customer_rank)
    _weight = (
        float(priority)  # type: ignore
        * _improvement  # type: ignore
        * _user_float_1  # type: ignore
        * _user_float_2  # type: ignore
        * _user_float_3  # type: ignore
        * _user_float_4  # type: ignore
        * _user_float_5  # type: ignore
    )

    return _improvement, _weight



# pylint: disable=unused-import, missing-docstring
# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.preferences.panel.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""RAMSTK GTK3 Preferences Panels."""

# Standard Library Imports
from typing import Any, Dict, List

# Third Party Imports
import toml
from pubsub import pub

# RAMSTK Package Imports
from ramstk.configuration import RAMSTKUserConfiguration
from ramstk.utilities import string_to_boolean
from ramstk.views.gtk3 import Gdk, Gtk, _
from ramstk.views.gtk3.widgets import (
    RAMSTKComboBox,
    RAMSTKEntry,
    RAMSTKFileChooserButton,
    RAMSTKFixedPanel,
    RAMSTKLabel,
    RAMSTKTreePanel,
    RAMSTKTreeView,
)


class GeneralPreferencesPanel(RAMSTKFixedPanel):
    """The panel to display options to be edited."""

    # Define private dictionary class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _select_msg = "succeed_get_preferences_attributes"
    _tag = "preferences"
    _title = _("General Preferences")

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self) -> None:
        """Initialize an instance of the Preferences panel."""
        super().__init__()

        # Initialize widgets.
        self.btnConfDir: RAMSTKFileChooserButton = RAMSTKFileChooserButton(
            _("RAMSTK Configuration File Directory")
        )
        self.btnDataDir: RAMSTKFileChooserButton = RAMSTKFileChooserButton(
            _("RAMSTK Data Directory")
        )
        self.btnIconDir: RAMSTKFileChooserButton = RAMSTKFileChooserButton(
            _("RAMSTK Icon Directory")
        )
        self.btnLogDir: RAMSTKFileChooserButton = RAMSTKFileChooserButton(
            _("RAMSTK Log Directory")
        )

        self.cmbModuleBookTabPosition: RAMSTKComboBox = RAMSTKComboBox(simple=True)
        self.cmbWorkBookTabPosition: RAMSTKComboBox = RAMSTKComboBox(simple=True)
        self.cmbListBookTabPosition: RAMSTKComboBox = RAMSTKComboBox(simple=True)
        self.cmbReportSize: RAMSTKComboBox = RAMSTKComboBox(simple=True)

        self.txtFRMultiplier: RAMSTKEntry = RAMSTKEntry()
        self.txtDecimalPlaces: RAMSTKEntry = RAMSTKEntry()
        self.txtMissionTime: RAMSTKEntry = RAMSTKEntry()

        # Initialize private dict instance attributes.

        # Initialize private list instance attributes.

        # Initialize private scalar instance attributes.
        self._configuration: RAMSTKUserConfiguration = RAMSTKUserConfiguration()

        # Initialize public dict instance attributes.
        self.dic_attribute_widget_map: Dict[str, List[Any]] = {
            "module_book_tab_pos": [
                0,
                self.cmbModuleBookTabPosition,
                "changed",
                self._on_changed_combo,
                "",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Module Book Tab Position:"),
                "gchararray",
            ],
            "work_book_tab_pos": [
                1,
                self.cmbWorkBookTabPosition,
                "changed",
                self._on_changed_combo,
                "",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Work Book Tab Position:"),
                "gchararray",
            ],
            "list_book_tab_pos": [
                2,
                self.cmbListBookTabPosition,
                "changed",
                self._on_changed_combo,
                "",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("List Book Tab Position:"),
                "gchararray",
            ],
            "report_size": [
                3,
                self.cmbReportSize,
                "changed",
                super().on_changed_combo,
                "",
                "Letter",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Report Paper Size:"),
                "gchararray",
            ],
            "fr_multiplier": [
                4,
                self.txtFRMultiplier,
                "changed",
                super().on_changed_entry,
                "",
                6,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                    "width": 75,
                },
                _("Failure Rate Multiplier:"),
                "gfloat",
            ],
            "decimals": [
                5,
                self.txtDecimalPlaces,
                "changed",
                super().on_changed_entry,
                "",
                3,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                    "width": 75,
                },
                _("Decimal Places:"),
                "gint",
            ],
            "mission_time": [
                6,
                self.txtMissionTime,
                "changed",
                super().on_changed_entry,
                "",
                1.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Reliability Mission Time:"),
                "gfloat",
            ],
            "config_file_path": [
                7,
                self.btnConfDir,
                "file-set",
                self._do_select_path,
                "",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "height": 30,
                    "fg_color": "#000000",
                    "select-action": Gtk.FileChooserAction.SELECT_FOLDER,
                    "visible": True,
                },
                _("Path to RAMSTK Configuration Files:"),
                "gchararray",
            ],
            "data_file_path": [
                8,
                self.btnDataDir,
                "file-set",
                self._do_select_path,
                "",
                1,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "height": 30,
                    "fg_color": "#000000",
                    "select-action": Gtk.FileChooserAction.SELECT_FOLDER,
                    "visible": True,
                },
                _("Path to RAMSTK Data Files:"),
                "gchararray",
            ],
            "icon_file_path": [
                9,
                self.btnIconDir,
                "file-set",
                self._do_select_path,
                "",
                1,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "height": 30,
                    "fg_color": "#000000",
                    "select-action": Gtk.FileChooserAction.SELECT_FOLDER,
                    "visible": True,
                },
                _("Path to RAMSTK Icon Files:"),
                "gchararray",
            ],
            "log_file_path": [
                10,
                self.btnLogDir,
                "file-set",
                self._do_select_path,
                "",
                2,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "height": 30,
                    "fg_color": "#000000",
                    "select-action": Gtk.FileChooserAction.SELECT_FOLDER,
                    "visible": True,
                },
                _("Path to RAMSTK Log Files:"),
                "gchararray",
            ],
        }

        # Initialize public list instance attributes.

        # Initialize public scalar instance attributes.

        super().do_set_properties()
        super().do_make_panel()
        self._do_load_comboboxes()
        super().do_set_callbacks()

        # Subscribe to PyPubSub messages.
        pub.subscribe(self._do_load_panel, "request_load_preferences")

    def _do_load_panel(self, configuration: RAMSTKUserConfiguration) -> None:
        """Load the current preference values.

        :return: None
        :rtype: None
        """
        _positions = {"bottom": 1, "left": 2, "right": 3, "top": 4}
        _papersize = {"a4": 1, "letter": 2}

        self._configuration = configuration

        self.cmbModuleBookTabPosition.do_update(
            _positions[self._configuration.RAMSTK_TABPOS["modulebook"].lower()],
            signal="changed",
        )
        self.cmbWorkBookTabPosition.do_update(
            _positions[self._configuration.RAMSTK_TABPOS["workbook"].lower()],
            signal="changed",
        )
        self.cmbListBookTabPosition.do_update(
            _positions[self._configuration.RAMSTK_TABPOS["listbook"].lower()],
            signal="changed",
        )
        self.cmbReportSize.do_update(
            _papersize[self._configuration.RAMSTK_REPORT_SIZE.lower()], signal="changed"
        )

        self.txtFRMultiplier.do_update(
            str(self._configuration.RAMSTK_HR_MULTIPLIER), signal="changed"
        )
        self.txtDecimalPlaces.do_update(
            str(self._configuration.RAMSTK_DEC_PLACES), signal="changed"
        )
        self.txtMissionTime.do_update(
            str(self._configuration.RAMSTK_MTIME), signal="changed"
        )

        self.btnConfDir.set_current_folder(self._configuration.RAMSTK_CONF_DIR)
        self.btnDataDir.set_current_folder(self._configuration.RAMSTK_DATA_DIR)
        self.btnIconDir.set_current_folder(self._configuration.RAMSTK_ICON_DIR)
        self.btnLogDir.set_current_folder(self._configuration.RAMSTK_LOG_DIR)

    def _do_load_comboboxes(self) -> None:
        """Load the RAMSTKComboBoxes() with their data.

        :return: None
        :rtype: None
        """
        self.cmbModuleBookTabPosition.do_load_combo(
            [["Bottom"], ["Left"], ["Right"], ["Top"]]
        )
        self.cmbWorkBookTabPosition.do_load_combo(
            [["Bottom"], ["Left"], ["Right"], ["Top"]]
        )
        self.cmbListBookTabPosition.do_load_combo(
            [["Bottom"], ["Left"], ["Right"], ["Top"]]
        )
        self.cmbReportSize.do_load_combo([["A4"], ["Letter"]])

    def _do_select_path(self, button: Gtk.FileChooserButton, index: int) -> None:
        """Select the path from the file chooser.

        :param button: the Gtk.FileChooserButton() that called this method.
        :param index: the index of the Gtk.FileChooserButton() that called
            this method.
        :return: None
        :rtyp: None
        """
        if index == 0:
            self._configuration.RAMSTK_CONF_DIR = button.get_current_folder()
        elif index == 1:
            self._configuration.RAMSTK_DATA_DIR = button.get_current_folder()
        elif index == 2:
            self._configuration.RAMSTK_ICON_DIR = button.get_current_folder()
        elif index == 3:
            self._configuration.RAMSTK_LOG_DIR = button.get_current_folder()

    def _on_changed_combo(self, combo: RAMSTKComboBox, index: int) -> None:
        """Edit RAMSTKTreeView() layouts.

        :param combo: the RAMSTKComboBox() that called this method.
        :type combo: :class:`gui.gtk.RAMSTKComboBox`
        :param index: the index in the signal handler list associated with
            the RAMSTKComboBox() calling this method.
        :return: None
        :rtype: None
        """
        combo.handler_block(combo.dic_handler_id["changed"])

        if index == 1:
            self._configuration.RAMSTK_TABPOS["modulebook"] = combo.get_value()
        elif index == 2:
            self._configuration.RAMSTK_TABPOS["workbook"] = combo.get_value()
        elif index == 3:
            self._configuration.RAMSTK_TABPOS["listbook"] = combo.get_value()
        elif index == 4:
            self._configuration.RAMSTK_REPORT_SIZE = combo.get_value()

        combo.handler_unblock(combo.dic_handler_id["changed"])


class LookFeelPreferencesPanel(RAMSTKFixedPanel):
    """The panel to display options to be edited."""

    # Define private dictionary class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _select_msg = "succeed_get_preferences_attributes"
    _tag = "preferences"
    _title = _("Look &amp; Feel")

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self) -> None:
        """Initialize an instance of the Look and Feel panel."""
        super().__init__()

        # Initialize widgets.
        self.btnRevisionBGColor = Gtk.ColorButton()
        self.btnRevisionFGColor = Gtk.ColorButton()
        self.btnFunctionBGColor = Gtk.ColorButton()
        self.btnFunctionFGColor = Gtk.ColorButton()
        self.btnRequirementsBGColor = Gtk.ColorButton()
        self.btnRequirementsFGColor = Gtk.ColorButton()
        self.btnHardwareBGColor = Gtk.ColorButton()
        self.btnHardwareFGColor = Gtk.ColorButton()
        self.btnValidationBGColor = Gtk.ColorButton()
        self.btnValidationFGColor = Gtk.ColorButton()

        # Initialize private dict instance attributes.

        # Initialize private list instance attributes.

        # Initialize private scalar instance attributes.
        self._configuration: RAMSTKUserConfiguration = RAMSTKUserConfiguration()

        # Initialize public dict instance attributes.
        self.dic_attribute_widget_map: Dict[str, List[Any]] = {
            "revisionbg": [
                0,
                self.btnRevisionBGColor,
                "color-set",
                self._do_set_color,
                "",
                "#FFFFFF",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "height": 30,
                    "visible": True,
                },
                _("Revision Tree Background Color:"),
            ],
            "revisionfg": [
                1,
                self.btnRevisionFGColor,
                "color-set",
                self._do_set_color,
                "",
                "#000000",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "height": 30,
                    "visible": True,
                },
                _("Revision Tree Foreground Color:"),
            ],
            "functionbg": [
                2,
                self.btnFunctionBGColor,
                "color-set",
                self._do_set_color,
                "",
                "#FFFFFF",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "height": 30,
                    "visible": True,
                },
                _("Function Tree Background Color:"),
            ],
            "functionfg": [
                3,
                self.btnFunctionFGColor,
                "color-set",
                self._do_set_color,
                "",
                "#000000",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "height": 30,
                    "visible": True,
                },
                _("Function Tree Foreground Color:"),
            ],
            "requirementbg": [
                4,
                self.btnRequirementsBGColor,
                "color-set",
                self._do_set_color,
                "",
                "#FFFFFF",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "height": 30,
                    "visible": True,
                    "width": 75,
                },
                _("Requirements Tree Background Color:"),
            ],
            "requirementfg": [
                5,
                self.btnRequirementsFGColor,
                "color-set",
                self._do_set_color,
                "",
                "#000000",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "height": 30,
                    "visible": True,
                    "width": 75,
                },
                _("Requirements Tree Foreground Color:"),
            ],
            "hardwarebg": [
                6,
                self.btnHardwareBGColor,
                "color-set",
                self._do_set_color,
                "",
                "#FFFFFF",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "height": 30,
                    "visible": True,
                },
                _("Hardware Tree Background Color:"),
            ],
            "hardwarefg": [
                7,
                self.btnHardwareFGColor,
                "color-set",
                self._do_set_color,
                "",
                "#000000",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "height": 30,
                    "visible": True,
                },
                _("Hardware Tree Foreground Color:"),
            ],
            "validationbg": [
                8,
                self.btnValidationBGColor,
                "color-set",
                self._do_set_color,
                "",
                "#FFFFFF",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "height": 30,
                    "visible": True,
                },
                _("Validation Tree Background Color:"),
            ],
            "validationfg": [
                9,
                self.btnValidationFGColor,
                "color-set",
                self._do_set_color,
                "",
                "#000000",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "height": 30,
                    "visible": True,
                },
                _("Validation Tree Foreground Color:"),
            ],
        }

        # Initialize public list instance attributes.

        # Initialize public scalar instance attributes.

        self._do_set_properties()
        super().do_make_panel()
        self._do_set_callbacks()

        # Subscribe to PyPubSub messages.
        pub.subscribe(self._do_load_panel, "request_load_preferences")

    def _do_load_panel(self, configuration: RAMSTKUserConfiguration) -> None:
        """Load the current preference values.

        :return: None
        :rtype: None
        """
        self._configuration = configuration

        self.btnRevisionBGColor.set_color(
            Gdk.color_parse(self._configuration.RAMSTK_COLORS["revisionbg"])
        )
        self.btnRevisionFGColor.set_color(
            Gdk.color_parse(self._configuration.RAMSTK_COLORS["revisionfg"])
        )
        self.btnFunctionBGColor.set_color(
            Gdk.color_parse(self._configuration.RAMSTK_COLORS["functionbg"])
        )
        self.btnFunctionFGColor.set_color(
            Gdk.color_parse(self._configuration.RAMSTK_COLORS["functionfg"])
        )
        self.btnRequirementsBGColor.set_color(
            Gdk.color_parse(self._configuration.RAMSTK_COLORS["requirementbg"])
        )
        self.btnRequirementsFGColor.set_color(
            Gdk.color_parse(self._configuration.RAMSTK_COLORS["requirementfg"])
        )
        self.btnHardwareBGColor.set_color(
            Gdk.color_parse(self._configuration.RAMSTK_COLORS["hardwarebg"])
        )
        self.btnHardwareFGColor.set_color(
            Gdk.color_parse(self._configuration.RAMSTK_COLORS["hardwarefg"])
        )
        self.btnValidationBGColor.set_color(
            Gdk.color_parse(self._configuration.RAMSTK_COLORS["validationbg"])
        )
        self.btnValidationFGColor.set_color(
            Gdk.color_parse(self._configuration.RAMSTK_COLORS["validationfg"])
        )

    def _do_set_color(self, colorbutton: Gtk.ColorButton, ramstk_color: int) -> None:
        """Set the selected color.

        :param colorbutton: the Gtk.ColorButton() that called this method.
        :param ramstk_color: the name of the color to set.
        :return: False if successful or True if an error is encountered.
        :rtype: bool
        """
        # Retrieve the six digit hexidecimal version of the selected color.
        _color = colorbutton.get_color()
        try:
            _red = f"{int(_color.red / 255)}:#02"
        except ValueError:
            _red = f"{int(_color.red / 255)}"
        try:
            _green = f"{int(_color.green / 255)}:#02"
        except ValueError:
            _green = f"{int(_color.green / 255)}"
        try:
            _blue = f"{int(_color.blue / 255)}:#02"
        except ValueError:
            _blue = f"{int(_color.blue / 255)}"
        _color = f"#{_red}{_green}{_blue}"

        # Set the color variable.
        self._configuration.RAMSTK_COLORS[ramstk_color] = _color

    def _do_set_callbacks(self) -> None:
        """Set the callback functions/methods for each of the widgets.

        :return: None
        :rtype: None
        """
        # ----- BUTTONS
        self.btnRevisionBGColor.connect("color-set", self._do_set_color, "revisionbg")
        self.btnRevisionFGColor.connect("color-set", self._do_set_color, "revisionfg")
        self.btnFunctionBGColor.connect("color-set", self._do_set_color, "functionbg")
        self.btnFunctionFGColor.connect("color-set", self._do_set_color, "functionfg")
        self.btnRequirementsBGColor.connect(
            "color-set", self._do_set_color, "requirementbg"
        )
        self.btnRequirementsFGColor.connect(
            "color-set", self._do_set_color, "requirementfg"
        )
        self.btnHardwareBGColor.connect("color-set", self._do_set_color, "hardwarebg")
        self.btnHardwareFGColor.connect("color-set", self._do_set_color, "hardwarefg")
        self.btnValidationBGColor.connect(
            "color-set", self._do_set_color, "validationbg"
        )
        self.btnValidationFGColor.connect(
            "color-set", self._do_set_color, "validationfg"
        )

    def _do_set_properties(self) -> None:
        """Set the properties of the Preferences assistance widgets.

        :return: None
        :rtype: None
        """
        # ----- BUTTONS
        self.btnRevisionBGColor.height = 30
        self.btnRevisionFGColor.height = 30
        self.btnFunctionBGColor.height = 30
        self.btnFunctionFGColor.height = 30
        self.btnRequirementsBGColor.height = 30
        self.btnRequirementsFGColor.height = 30
        self.btnHardwareBGColor.height = 30
        self.btnHardwareFGColor.height = 30
        self.btnValidationBGColor.height = 30
        self.btnValidationFGColor.height = 30


class TreeLayoutPreferencesPanel(RAMSTKTreePanel):
    """The panel to display options to be edited."""

    # Define private dictionary class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _select_msg = None
    _tag = "preferences"
    _title = _("Tree View Layout")

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self) -> None:
        """Initialize an instance of the RAMSTKTreeView Layout panel."""
        super().__init__()

        # Initialize widgets.
        self.cmbFormatFiles: RAMSTKComboBox = RAMSTKComboBox(simple=False)
        self.tvwTreeView: RAMSTKTreeView = RAMSTKTreeView()

        # Initialize private dict instance attributes.

        # Initialize private list instance attributes.

        # Initialize private scalar instance attributes.
        self._configuration: RAMSTKUserConfiguration = RAMSTKUserConfiguration()

        # Initialize public dict instance attributes.

        # Initialize public list instance attributes.

        # Initialize public scalar instance attributes.
        self.fmt_file: str = ""

        # Make a fixed type panel.
        self._do_load_comboboxes()
        self.__do_make_treeview()
        self.__make_ui()

        self.cmbFormatFiles.do_set_properties(
            tooltip=_("Select the Tree View layout to edit.")
        )
        self.cmbFormatFiles.dic_handler_id["changed"] = self.cmbFormatFiles.connect(
            "changed", self._on_combo_changed
        )

        # Subscribe to PyPubSub messages.
        pub.subscribe(self._do_load_panel, "request_load_preferences")

    def _do_load_comboboxes(self) -> None:
        """Load the RAMSTKComboBoxes() with their data.

        :return: None
        :rtype: None
        """
        self.cmbFormatFiles.do_load_combo(
            [
                [_("Allocation"), "allocation", ""],
                [_("Failure Definition"), "failure_definition", ""],
                [_("(D)FME(C)A"), "fmea", ""],
                [_("Function"), "function", ""],
                [_("Hardware"), "hardware", ""],
                [_("Hazards Analysis"), "hazard", ""],
                [_("Physics of Failure Analysis"), "pof", ""],
                [_("Requirements"), "requirement", ""],
                [_("Revisions"), "revision", ""],
                [_("Similar Item Analysis"), "similar_item", ""],
                [_("Stakeholder Inputs"), "stakeholder", ""],
                [_("Usage Profile"), "usage_profile", ""],
                [_("Validation"), "validation", ""],
            ],
            simple=False,
        )

    def _do_load_format(self, module: str) -> None:
        """Load the selected Module View format file for editing.

        :param module: the name of the RAMSTK workstream module whose
            Module View layout is to be edited.
        :return: None
        :rtype: None
        """
        self.fmt_file = (
            self._configuration.RAMSTK_CONF_DIR
            + "/layouts/"
            + self._configuration.RAMSTK_FORMAT_FILE[module]
        )

        _format = toml.load(self.fmt_file)

        _datatypes = _format["datatype"]
        _defaulttitle = _format["defaulttitle"]
        _editable = _format["editable"]
        _usertitle = _format["usertitle"]
        _keys = _format["key"]
        _position = _format["position"]
        _visible = _format["visible"]
        _widgets = _format["widget"]

        _model = self.tvwTreeView.get_model()
        _model.clear()
        for _key in _defaulttitle:
            _data = [
                _defaulttitle[_key],
                _usertitle[_key],
                int(_position[_key]),
                string_to_boolean(_editable[_key]),
                string_to_boolean(_visible[_key]),
                _datatypes[_key],
                _widgets[_key],
                _keys[_key],
                _key,
            ]
            _model.append(None, _data)

    def _do_load_panel(self, configuration: RAMSTKUserConfiguration) -> None:
        """Load the current preference values.

        :return: None
        :rtype: None
        """
        self._configuration = configuration

    def _on_combo_changed(self, combo: RAMSTKComboBox) -> None:
        """Edit RAMSTKTreeView() layouts.

        :param combo: the RAMSTKComboBox() that called this method.
        :return: None
        :rtype: None
        """
        combo.handler_block(combo.dic_handler_id["changed"])

        _model = combo.get_model()
        _row = combo.get_active_iter()
        _module = _model.get_value(_row, 1)
        self._do_load_format(_module)

        combo.handler_unblock(combo.dic_handler_id["changed"])

    def __do_make_treeview(self) -> None:
        """Make the format file editing Gtk.Treeview().

        :return: None
        :rtype: None
        """
        self.tvwTreeView.position = {
            "col0": 0,
            "col1": 1,
            "col2": 2,
            "col3": 3,
            "col4": 4,
            "col5": 5,
            "col6": 6,
            "col7": 7,
            "col8": 8,
        }
        self.tvwTreeView.headings = {
            "col0": _("Default\nTitle"),
            "col1": _("User\nTitle"),
            "col2": _("Column\nPosition"),
            "col3": _("Can\nEdit?"),
            "col4": _("Is\nVisible?"),
            "col5": "",
            "col6": "",
            "col7": "",
            "col8": "",
        }
        self.tvwTreeView.editable = {
            "col0": "False",
            "col1": "True",
            "col2": "True",
            "col3": "True",
            "col4": "True",
            "col5": "False",
            "col6": "False",
            "col7": "False",
            "col8": "False",
        }
        self.tvwTreeView.visible = {
            "col0": "True",
            "col1": "True",
            "col2": "True",
            "col3": "True",
            "col4": "True",
            "col5": "False",
            "col6": "False",
            "col7": "False",
            "col8": "False",
        }
        self.tvwTreeView.datatypes = {
            "col0": "gchararray",
            "col1": "gchararray",
            "col2": "gint",
            "col3": "gint",
            "col4": "gint",
            "col5": "gchararray",
            "col6": "gchararray",
            "col7": "gchararray",
            "col8": "gchararray",
        }
        self.tvwTreeView.korder = {
            "col0": "default_title",
            "col1": "user_title",
            "col2": "column_position",
            "col3": "can_edit",
            "col4": "is_visible",
            "col5": "unk1",
            "col6": "unk2",
            "col7": "unk3",
            "col8": "unk4",
        }
        self.tvwTreeView.widgets = {
            "default_title": Gtk.CellRendererText(),
            "user_title": Gtk.CellRendererText(),
            "column_position": Gtk.CellRendererText(),
            "can_edit": Gtk.CellRendererToggle(),
            "is_visible": Gtk.CellRendererToggle(),
            "unk1": Gtk.CellRendererText(),
            "unk2": Gtk.CellRendererText(),
            "unk3": Gtk.CellRendererText(),
            "unk4": Gtk.CellRendererText(),
        }
        self.tvwTreeView.do_make_model()
        self.tvwTreeView.do_make_columns()
        self.tvwTreeView.do_set_editable_columns(self.tvwTreeView.do_edit_cell)

    def __make_ui(self) -> None:
        """Build the UI for the Preferences assistant."""
        super().do_make_panel()

        _scrollwindow = self.get_child()
        self.remove(_scrollwindow)

        _label = RAMSTKLabel(_("Select format file to edit:"))
        _x_pos = _label.get_attribute("width")

        _fixed: Gtk.Fixed = Gtk.Fixed()
        _fixed.put(_label, 5, 5)
        _fixed.put(self.cmbFormatFiles, _x_pos + 10, 5)

        _vbox = Gtk.VBox()
        _vbox.pack_start(_fixed, False, False, 0)
        _vbox.pack_end(_scrollwindow, True, True, 0)

        self.add(_vbox)



# Standard Library Imports
from typing import Any

# RAMSTK Package Imports
from ramstk.configuration import RAMSTKUserConfiguration as RAMSTKUserConfiguration
from ramstk.views.gtk3 import Gtk as Gtk
from ramstk.views.gtk3 import _ as _
from ramstk.views.gtk3.widgets import RAMSTKFileChooser as RAMSTKFileChooser
from ramstk.views.gtk3.widgets import RAMSTKMessageDialog as RAMSTKMessageDialog

class ExportProject(RAMSTKFileChooser):
    RAMSTK_USER_CONFIGURATION: Any = ...
    _parent: Any = ...
    def __init__(
        self,
        __button: Gtk.ToolButton,
        configuration: RAMSTKUserConfiguration,
        parent: object,
    ) -> None: ...
    def _do_quit(self) -> None: ...
    def _do_select_file(self) -> None: ...



# pylint: disable=duplicate-code
# -*- coding: utf-8 -*-
#
#       ramstk.models.cause.record.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Cause Record Model."""

# Third Party Imports
from sqlalchemy import Column, ForeignKeyConstraint, Integer, String
from sqlalchemy.orm import relationship

# RAMSTK Package Imports
from ramstk.db import RAMSTK_BASE
from ramstk.models import RAMSTKBaseRecord


class RAMSTKCauseRecord(RAMSTK_BASE, RAMSTKBaseRecord):
    """Class to represent table ramstk_cause in the RAMSTK Program database.

    This table shares a Many-to-One relationship with ramstk_mechanism. This
    table shared a One-to-Many relationship with ramstk_control. This table
    shared a One-to-Many relationship with ramstk_action.
    """

    __defaults__ = {
        "description": "",
        "rpn": 0,
        "rpn_detection": 10,
        "rpn_detection_new": 10,
        "rpn_new": 0,
        "rpn_occurrence": 10,
        "rpn_occurrence_new": 10,
    }
    __tablename__ = "ramstk_cause"
    __table_args__ = (
        ForeignKeyConstraint(
            ["fld_revision_id", "fld_hardware_id", "fld_mode_id", "fld_mechanism_id"],
            [
                "ramstk_mechanism.fld_revision_id",
                "ramstk_mechanism.fld_hardware_id",
                "ramstk_mechanism.fld_mode_id",
                "ramstk_mechanism.fld_mechanism_id",
            ],
        ),
        {"extend_existing": True},
    )

    revision_id = Column("fld_revision_id", Integer, primary_key=True, nullable=False)
    hardware_id = Column(
        "fld_hardware_id", Integer, primary_key=True, default=-1, nullable=False
    )
    mode_id = Column("fld_mode_id", Integer, primary_key=True, nullable=False)
    mechanism_id = Column("fld_mechanism_id", Integer, primary_key=True, nullable=False)
    cause_id = Column(
        "fld_cause_id",
        Integer,
        primary_key=True,
        autoincrement=True,
        nullable=False,
        unique=True,
    )

    description = Column(
        "fld_description", String(512), default=__defaults__["description"]
    )
    rpn = Column("fld_rpn", Integer, default=__defaults__["rpn"])
    rpn_detection = Column(
        "fld_rpn_detection", Integer, default=__defaults__["rpn_detection"]
    )
    rpn_detection_new = Column(
        "fld_rpn_detection_new", Integer, default=__defaults__["rpn_detection_new"]
    )
    rpn_new = Column("fld_rpn_new", Integer, default=__defaults__["rpn_new"])
    rpn_occurrence = Column(
        "fld_rpn_occurrence", Integer, default=__defaults__["rpn_occurrence"]
    )
    rpn_occurrence_new = Column(
        "fld_rpn_occurrence_new", Integer, default=__defaults__["rpn_occurrence_new"]
    )

    # Define the relationships to other tables in the RAMSTK Program database.
    mechanism: relationship = relationship(
        "RAMSTKMechanismRecord", back_populates="cause"
    )
    control: relationship = relationship(
        "RAMSTKControlRecord", back_populates="cause", cascade="delete, delete-orphan"
    )
    action: relationship = relationship(
        "RAMSTKActionRecord", back_populates="cause", cascade="delete, delete-orphan"
    )

    is_mode = False
    is_mechanism = False
    is_cause = True
    is_control = False
    is_action = False

    def get_attributes(self):
        """Retrieve current values of the RAMSTKCause data model attributes.

        :return: {mode_id, mechanism_id, cause_id, description, rpn_occurrence,
                  rpn_detection, rpn, rpn_occurrence_new, rpn_detection_new,
                  rpn_new}
        :rtype: tuple
        """
        _attributes = {
            "mode_id": self.mode_id,
            "mechanism_id": self.mechanism_id,
            "cause_id": self.cause_id,
            "description": self.description,
            "rpn": self.rpn,
            "rpn_detection": self.rpn_detection,
            "rpn_detection_new": self.rpn_detection_new,
            "rpn_new": self.rpn_new,
            "rpn_occurrence": self.rpn_occurrence,
            "rpn_occurrence_new": self.rpn_occurrence_new,
        }

        return _attributes



# -*- coding: utf-8 -*-
#
#       ramstk.analyses.fha.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright 2019 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Functional Hazards Analysis (FHA) Module."""

# Standard Library Imports
from typing import Any, Dict, List

# Third Party Imports
# noinspection PyPackageRequirements
from sympy import symbols, sympify  # type: ignore

# RAMSTK Package Imports
from ramstk.exceptions import OutOfRangeError

PROBABILITY = {
    "Level E - Extremely Unlikely": 1,
    "Level D - Remote": 2,
    "Level C - Occasional": 3,
    "Level B - Reasonably Probable": 4,
    "Level A - Frequent": 5,
}
SEVERITY = {
    "Insignificant": 1,
    "Slight": 2,
    "Low": 3,
    "Medium": 4,
    "High": 5,
    "Major": 6,
}


def calculate_hri(probability: str, severity: str) -> int:
    """Calculate the hazard risk index (HRI).

    .. note:: See MIL-STD-882.

    :param probability: the hazard probability expressed in text.
    :param severity: the hazard severity expressed in text.
    :return: _hri; the calculated hazard risk index.
    :rtype: int
    :raise: OutOfRangeError if passed an unknown probability or severity
        description.
    """
    try:
        return PROBABILITY[probability] * SEVERITY[severity]
    except KeyError as _error:
        raise OutOfRangeError(
            (
                "calculate_hri() was passed an unknown hazard "
                "probability ({0:s}) or severity ({1:s}) "
                "description."
            ).format(probability, severity)
        ) from _error


def calculate_user_defined(fha: Dict[str, Any]) -> Dict[str, Any]:
    """Calculate the user-defined hazards analysis.

    :param fha: the user-defined functional hazards assessment dict.  The
        calling method/function should create the fha dict as follows:

        fha = OrderedDict({
            _key: None
            for _key in [
                'uf1', 'uf2', 'uf3', 'ui1', 'ui2', 'ui3', 'equation1',
                'equation2', 'equation3', 'equation4', 'equation5', 'res1',
                'res2', 'res3', 'res4', 'res5'
                ]
            })

    :return: fha; the functional hazards assessment dict with updated results.
    :rtype: dict
    """
    (uf1, uf2, uf3, ui1, ui2, ui3, res1, res2, res3, res4, res5) = symbols(
        "uf1 uf2 uf3 ui1 ui2 ui3 res1 res2 res3 res4 res5"
    )

    # pylint: disable=eval-used
    fha["res1"] = sympify(fha["equation1"]).evalf(
        subs={
            uf1: fha["uf1"],
            uf2: fha["uf2"],
            uf3: fha["uf3"],
            ui1: fha["ui1"],
            ui2: fha["ui2"],
            ui3: fha["ui3"],
            res1: fha["res1"],
            res2: fha["res2"],
            res3: fha["res3"],
            res4: fha["res4"],
            res5: fha["res5"],
        }
    )
    fha["res2"] = sympify(fha["equation2"]).evalf(
        subs={
            uf1: fha["uf1"],
            uf2: fha["uf2"],
            uf3: fha["uf3"],
            ui1: fha["ui1"],
            ui2: fha["ui2"],
            ui3: fha["ui3"],
            res1: fha["res1"],
            res2: fha["res2"],
            res3: fha["res3"],
            res4: fha["res4"],
            res5: fha["res5"],
        }
    )
    fha["res3"] = sympify(fha["equation3"]).evalf(
        subs={
            uf1: fha["uf1"],
            uf2: fha["uf2"],
            uf3: fha["uf3"],
            ui1: fha["ui1"],
            ui2: fha["ui2"],
            ui3: fha["ui3"],
            res1: fha["res1"],
            res2: fha["res2"],
            res3: fha["res3"],
            res4: fha["res4"],
            res5: fha["res5"],
        }
    )
    fha["res4"] = sympify(fha["equation4"]).evalf(
        subs={
            uf1: fha["uf1"],
            uf2: fha["uf2"],
            uf3: fha["uf3"],
            ui1: fha["ui1"],
            ui2: fha["ui2"],
            ui3: fha["ui3"],
            res1: fha["res1"],
            res2: fha["res2"],
            res3: fha["res3"],
            res4: fha["res4"],
            res5: fha["res5"],
        }
    )
    fha["res5"] = sympify(fha["equation5"]).evalf(
        subs={
            uf1: fha["uf1"],
            uf2: fha["uf2"],
            uf3: fha["uf3"],
            ui1: fha["ui1"],
            ui2: fha["ui2"],
            ui3: fha["ui3"],
            res1: fha["res1"],
            res2: fha["res2"],
            res3: fha["res3"],
            res4: fha["res4"],
            res5: fha["res5"],
        }
    )

    return fha


def set_user_defined_floats(fha: Dict[str, Any], floats: List[float]) -> Dict[str, Any]:
    """Set the user-defined float values for the user-defined calculations.

    :param fha: the functional hazard assessment dict.
    :param list floats: the list of float values.
    :return: fha; the functional hazard assessment dict with updated float
        values.
    :rtype: dict
    """
    _key = ""
    for _idx in [0, 1, 2]:
        try:
            _key = list(fha.keys())[_idx]
            fha[_key] = float(floats[_idx])
        except IndexError:
            fha[_key] = 0.0

    return fha


def set_user_defined_ints(fha: Dict[str, Any], ints: List[int]) -> Dict[str, Any]:
    """Set the user-defined integer values for the user-defined calculations.

    :param fha: the functional hazard assessment dict.
    :param list ints: the list of integer values.
    :return: fha; the functional hazard assessment dict with updated integer
        values.
    :rtype: dict
    """
    _key = ""
    for _idx in [3, 4, 5]:
        try:
            _key = list(fha.keys())[_idx]
            fha[_key] = int(ints[_idx - 3])
        except IndexError:
            fha[_key] = 0

    return fha


def set_user_defined_functions(
    fha: Dict[str, Any], functions: List[str]
) -> Dict[str, Any]:
    """Set the user-defined functions for the user-defined calculations.

    .. note:: by default we set the function equal to 0.0.  This prevents Sympy
        errors resulting from empty strings.

    :param fha: the functional hazard assessment dict.
    :param list functions: the list of functions; list items are str.
    :return: fha; the functional hazard assessment dict with updated functions.
    :rtype: dict
    """
    _key = ""
    for _idx in [6, 7, 8, 9, 10]:
        try:
            _key = list(fha.keys())[_idx]
            if str(functions[_idx - 6]) == "":
                fha[_key] = "0.0"
            else:
                fha[_key] = str(functions[_idx - 6])
        except IndexError:
            fha[_key] = "0.0"

    return fha


def set_user_defined_results(
    fha: Dict[str, Any], results: List[float]
) -> Dict[str, Any]:
    """Set the user-defined results for the user-defined calculations.

    This allows the use of the results fields to be manually set to float
    values by the user essentially creating five more user-defined float
    values.

    :param fha: the functional hazard assessment dict.
    :param list results: the list of results.
    :return: fha; the functional hazard assessment dict with updated results.
    :rtype: dict
    """
    _key = ""
    for _idx in [11, 12, 13, 14, 15]:
        try:
            _key = list(fha.keys())[_idx]
            fha[_key] = results[_idx - 11]
        except IndexError:
            fha[_key] = 0

    return fha



# pylint: disable=duplicate-code
# -*- coding: utf-8 -*-
#
#       ramstk.models.programdb.hazard.record.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Hazard Package Record Model."""

# Standard Library Imports
from typing import Any, Dict

# Third Party Imports
from sqlalchemy import Column, Float, ForeignKey, Integer, String

# RAMSTK Package Imports
from ramstk.db import RAMSTK_BASE
from ramstk.models import RAMSTKBaseRecord


class RAMSTKHazardRecord(RAMSTK_BASE, RAMSTKBaseRecord):
    """Class to represent ramstk_hazard_analysis table in the Program database.

    This table shares a Many-to-One relationship with ramstk_revision. This table
    shares a Many-to-One relationship with ramstk_function.
    """

    __defaults__ = {
        "potential_hazard": "",
        "potential_cause": "",
        "assembly_effect": "",
        "assembly_severity": "Major",
        "assembly_probability": "Level A - Frequent",
        "assembly_hri": 20,
        "assembly_mitigation": "",
        "assembly_severity_f": "Major",
        "assembly_probability_f": "Level A - Frequent",
        "assembly_hri_f": 20,
        "function_1": "",
        "function_2": "",
        "function_3": "",
        "function_4": "",
        "function_5": "",
        "remarks": "",
        "result_1": 0.0,
        "result_2": 0.0,
        "result_3": 0.0,
        "result_4": 0.0,
        "result_5": 0.0,
        "system_effect": "",
        "system_severity": "Major",
        "system_probability": "Level A - Frequent",
        "system_hri": 20,
        "system_mitigation": "",
        "system_severity_f": "Major",
        "system_probability_f": "Level A - Frequent",
        "system_hri_f": 20,
        "user_blob_1": "",
        "user_blob_2": "",
        "user_blob_3": "",
        "user_float_1": 0.0,
        "user_float_2": 0.0,
        "user_float_3": 0.0,
        "user_int_1": 0,
        "user_int_2": 0,
        "user_int_3": 0,
    }
    __tablename__ = "ramstk_hazard_analysis"
    __table_args__ = {"extend_existing": True}

    revision_id = Column(
        "fld_revision_id",
        Integer,
        ForeignKey("ramstk_revision.fld_revision_id", ondelete="CASCADE"),
        nullable=False,
    )
    function_id = Column(
        "fld_function_id",
        Integer,
        ForeignKey("ramstk_function.fld_function_id", ondelete="CASCADE"),
        nullable=False,
    )
    hazard_id = Column(
        "fld_hazard_id",
        Integer,
        primary_key=True,
        autoincrement=True,
        nullable=False,
        info={"identity": (0, 1)},
    )

    potential_hazard = Column(
        "fld_potential_hazard", String(256), default=__defaults__["potential_hazard"]
    )
    potential_cause = Column(
        "fld_potential_cause", String(512), default=__defaults__["potential_cause"]
    )
    assembly_effect = Column(
        "fld_assembly_effect", String(512), default=__defaults__["assembly_effect"]
    )
    assembly_severity = Column(
        "fld_assembly_severity", String(256), default=__defaults__["assembly_severity"]
    )
    assembly_probability = Column(
        "fld_assembly_probability",
        String(256),
        default=__defaults__["assembly_probability"],
    )
    assembly_hri = Column(
        "fld_assembly_hri", Integer, default=__defaults__["assembly_hri"]
    )
    assembly_mitigation = Column(
        "fld_assembly_mitigation", String, default=__defaults__["assembly_mitigation"]
    )
    assembly_severity_f = Column(
        "fld_assembly_severity_f",
        String(256),
        default=__defaults__["assembly_severity_f"],
    )
    assembly_probability_f = Column(
        "fld_assembly_probability_f",
        String(256),
        default=__defaults__["assembly_probability_f"],
    )
    assembly_hri_f = Column(
        "fld_assembly_hri_f", Integer, default=__defaults__["assembly_hri_f"]
    )
    function_1 = Column(
        "fld_function_1", String(128), default=__defaults__["function_1"]
    )
    function_2 = Column(
        "fld_function_2", String(128), default=__defaults__["function_2"]
    )
    function_3 = Column(
        "fld_function_3", String(128), default=__defaults__["function_3"]
    )
    function_4 = Column(
        "fld_function_4", String(128), default=__defaults__["function_4"]
    )
    function_5 = Column(
        "fld_function_5", String(128), default=__defaults__["function_5"]
    )
    remarks = Column("fld_remarks", String, default=__defaults__["remarks"])
    result_1 = Column("fld_result_1", Float, default=__defaults__["result_1"])
    result_2 = Column("fld_result_2", Float, default=__defaults__["result_2"])
    result_3 = Column("fld_result_3", Float, default=__defaults__["result_3"])
    result_4 = Column("fld_result_4", Float, default=__defaults__["result_4"])
    result_5 = Column("fld_result_5", Float, default=__defaults__["result_5"])
    system_effect = Column(
        "fld_system_effect", String(512), default=__defaults__["system_effect"]
    )
    system_severity = Column(
        "fld_system_severity", String(256), default=__defaults__["system_severity"]
    )
    system_probability = Column(
        "fld_system_probability",
        String(256),
        default=__defaults__["system_probability"],
    )
    system_hri = Column("fld_system_hri", Integer, default=__defaults__["system_hri"])
    system_mitigation = Column(
        "fld_system_mitigation", String, default=__defaults__["system_mitigation"]
    )
    system_severity_f = Column(
        "fld_system_severity_f", String(256), default=__defaults__["system_severity_f"]
    )
    system_probability_f = Column(
        "fld_system_probability_f",
        String(256),
        default=__defaults__["system_probability_f"],
    )
    system_hri_f = Column(
        "fld_system_hri_f", Integer, default=__defaults__["system_hri_f"]
    )
    user_blob_1 = Column("fld_user_blob_1", String, default=__defaults__["user_blob_1"])
    user_blob_2 = Column("fld_user_blob_2", String, default=__defaults__["user_blob_2"])
    user_blob_3 = Column("fld_user_blob_3", String, default=__defaults__["user_blob_3"])
    user_float_1 = Column(
        "fld_user_float_1", Float, default=__defaults__["user_float_1"]
    )
    user_float_2 = Column(
        "fld_user_float_2", Float, default=__defaults__["user_float_2"]
    )
    user_float_3 = Column(
        "fld_user_float_3", Float, default=__defaults__["user_float_3"]
    )
    user_int_1 = Column("fld_user_int_1", Integer, default=__defaults__["user_int_1"])
    user_int_2 = Column("fld_user_int_2", Integer, default=__defaults__["user_int_2"])
    user_int_3 = Column("fld_user_int_3", Integer, default=__defaults__["user_int_3"])

    # Define the relationships to other tables in the RAMSTK Program database.

    def get_attributes(self) -> Dict[str, Any]:
        """Retrieve current values of RAMSTKHazardAnalysis model attributes.

        :return: {revision_id, hardware_id, hazard_id, potential_hazard,
                  potential_cause, assembly_effect, assembly_severity_id,
                  assembly_probability_id, assembly_hri, assembly_mitigation,
                  assembly_severity_id_f, assembly_probability_id_f,
                  assembly_hri_f, system_effect, system_severity_id,
                  system_probability_id, system_hri, system_mitigation,
                  system_severity_id_f, system_probability_id_f,
                  system_hri_f, remarks, function_1, function_2, function_3,
                  function_4, function_5, result_1, result_2, result_3,
                  result_4, result_5, user_blob_1, user_blob_2, user_blob_3,
                  user_float_1, user_float_2, user_float_3, user_int_1,
                  user_int_2, user_int_3} pairs
        :rtype: dict
        """
        _attributes = {
            "revision_id": self.revision_id,
            "function_id": self.function_id,
            "hazard_id": self.hazard_id,
            "potential_hazard": self.potential_hazard,
            "potential_cause": self.potential_cause,
            "assembly_effect": self.assembly_effect,
            "assembly_severity": self.assembly_severity,
            "assembly_probability": self.assembly_probability,
            "assembly_hri": self.assembly_hri,
            "assembly_mitigation": self.assembly_mitigation,
            "assembly_severity_f": self.assembly_severity_f,
            "assembly_probability_f": self.assembly_probability_f,
            "assembly_hri_f": self.assembly_hri_f,
            "system_effect": self.system_effect,
            "system_severity": self.system_severity,
            "system_probability": self.system_probability,
            "system_hri": self.system_hri,
            "system_mitigation": self.system_mitigation,
            "system_severity_f": self.system_severity_f,
            "system_probability_f": self.system_probability_f,
            "system_hri_f": self.system_hri_f,
            "remarks": self.remarks,
            "function_1": self.function_1,
            "function_2": self.function_2,
            "function_3": self.function_3,
            "function_4": self.function_4,
            "function_5": self.function_5,
            "result_1": self.result_1,
            "result_2": self.result_2,
            "result_3": self.result_3,
            "result_4": self.result_4,
            "result_5": self.result_5,
            "user_blob_1": self.user_blob_1,
            "user_blob_2": self.user_blob_2,
            "user_blob_3": self.user_blob_3,
            "user_float_1": self.user_float_1,
            "user_float_2": self.user_float_2,
            "user_float_3": self.user_float_3,
            "user_int_1": self.user_int_1,
            "user_int_2": self.user_int_2,
            "user_int_3": self.user_int_3,
        }

        return _attributes



# pylint: disable=duplicate-code, too-many-instance-attributes
# -*- coding: utf-8 -*-
#
#       ramstk.models.design_mechanic.record.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""RAMSTKDesignMechanic Table Module."""

# Third Party Imports
from sqlalchemy import Column, Float, ForeignKey, Integer

# RAMSTK Package Imports
from ramstk.db import RAMSTK_BASE
from ramstk.models import RAMSTKBaseRecord


class RAMSTKDesignMechanicRecord(RAMSTK_BASE, RAMSTKBaseRecord):
    """Represent ramstk_design_mechanic table in the RAMSTK Program database.

    This table shares a One-to-One relationship with ramstk_hardware.
    """

    __defaults__ = {
        "altitude_operating": 0.0,
        "application_id": 0,
        "balance_id": 0,
        "clearance": 0.0,
        "casing_id": 0,
        "contact_pressure": 0.0,
        "deflection": 0.0,
        "diameter_coil": 0.0,
        "diameter_inner": 0.0,
        "diameter_outer": 0.0,
        "diameter_wire": 0.0,
        "filter_size": 0.0,
        "flow_design": 0.0,
        "flow_operating": 0.0,
        "frequency_operating": 0.0,
        "friction": 0.0,
        "impact_id": 0,
        "leakage_allowable": 0.0,
        "length": 0.0,
        "length_compressed": 0.0,
        "length_relaxed": 0.0,
        "load_design": 0.0,
        "load_id": 0,
        "load_operating": 0.0,
        "lubrication_id": 0,
        "manufacturing_id": 0,
        "material_id": 0,
        "meyer_hardness": 0.0,
        "misalignment_angle": 0.0,
        "n_ten": 0,
        "n_cycles": 0,
        "n_elements": 0,
        "offset": 0.0,
        "particle_size": 0.0,
        "pressure_contact": 0.0,
        "pressure_delta": 0.0,
        "pressure_downstream": 0.0,
        "pressure_rated": 0.0,
        "pressure_upstream": 0.0,
        "rpm_design": 0.0,
        "rpm_operating": 0.0,
        "service_id": 0,
        "spring_index": 0.0,
        "surface_finish": 0.0,
        "technology_id": 0,
        "thickness": 0.0,
        "torque_id": 0,
        "type_id": 0,
        "viscosity_design": 0.0,
        "viscosity_dynamic": 0.0,
        "water_per_cent": 0.0,
        "width_minimum": 0.0,
    }
    __tablename__ = "ramstk_design_mechanic"
    __table_args__ = {"extend_existing": True}

    revision_id = Column(
        "fld_revision_id",
        Integer,
        ForeignKey("ramstk_revision.fld_revision_id", ondelete="CASCADE"),
        nullable=False,
    )
    hardware_id = Column(
        "fld_hardware_id",
        Integer,
        ForeignKey("ramstk_hardware.fld_hardware_id", ondelete="CASCADE"),
        primary_key=True,
        nullable=False,
    )

    altitude_operating = Column(
        "fld_altitude_operating", Float, default=__defaults__["altitude_operating"]
    )
    application_id = Column(
        "fld_application_id", Integer, default=__defaults__["application_id"]
    )
    balance_id = Column("fld_balance_id", Integer, default=__defaults__["balance_id"])
    clearance = Column("fld_clearance", Float, default=__defaults__["clearance"])
    casing_id = Column("fld_casing_id", Integer, default=__defaults__["casing_id"])
    contact_pressure = Column(
        "fld_contact_pressure", Float, default=__defaults__["contact_pressure"]
    )
    deflection = Column("fld_deflection", Float, default=__defaults__["deflection"])
    diameter_coil = Column(
        "fld_diameter_coil", Float, default=__defaults__["diameter_coil"]
    )
    diameter_inner = Column(
        "fld_diameter_inner", Float, default=__defaults__["diameter_inner"]
    )
    diameter_outer = Column(
        "fld_diameter_outer", Float, default=__defaults__["diameter_outer"]
    )
    diameter_wire = Column(
        "fld_diameter_wire", Float, default=__defaults__["diameter_wire"]
    )
    filter_size = Column("fld_filter_size", Float, default=__defaults__["filter_size"])
    flow_design = Column("fld_flow_design", Float, default=__defaults__["flow_design"])
    flow_operating = Column(
        "fld_flow_operating", Float, default=__defaults__["flow_operating"]
    )
    frequency_operating = Column(
        "fld_frequency_operating", Float, default=__defaults__["frequency_operating"]
    )
    friction = Column("fld_friction", Float, default=__defaults__["friction"])
    impact_id = Column("fld_impact_id", Integer, default=__defaults__["impact_id"])
    leakage_allowable = Column(
        "fld_leakage_allowable", Float, default=__defaults__["leakage_allowable"]
    )
    length = Column("fld_length", Float, default=__defaults__["length"])
    length_compressed = Column(
        "fld_length_compressed", Float, default=__defaults__["length_compressed"]
    )
    length_relaxed = Column(
        "fld_length_relaxed", Float, default=__defaults__["length_relaxed"]
    )
    load_design = Column("fld_load_design", Float, default=__defaults__["load_design"])
    load_id = Column("fld_load_id", Integer, default=__defaults__["load_id"])
    load_operating = Column(
        "fld_load_operating", Float, default=__defaults__["load_operating"]
    )
    lubrication_id = Column(
        "fld_lubrication_id", Integer, default=__defaults__["lubrication_id"]
    )
    manufacturing_id = Column(
        "fld_manufacturing_id", Integer, default=__defaults__["manufacturing_id"]
    )
    material_id = Column(
        "fld_material_id", Integer, default=__defaults__["material_id"]
    )
    meyer_hardness = Column(
        "fld_meyer_hardness", Float, default=__defaults__["meyer_hardness"]
    )
    misalignment_angle = Column(
        "fld_misalignment_angle", Float, default=__defaults__["misalignment_angle"]
    )
    n_ten = Column("fld_n_ten", Integer, default=__defaults__["n_ten"])
    n_cycles = Column("fld_n_cycles", Integer, default=__defaults__["n_cycles"])
    n_elements = Column("fld_n_elements", Integer, default=__defaults__["n_elements"])
    offset = Column("fld_offset", Float, default=__defaults__["offset"])
    particle_size = Column(
        "fld_particle_size", Float, default=__defaults__["particle_size"]
    )
    pressure_contact = Column(
        "fld_pressure_contact", Float, default=__defaults__["pressure_contact"]
    )
    pressure_delta = Column(
        "fld_pressure_delta", Float, default=__defaults__["pressure_delta"]
    )
    pressure_downstream = Column(
        "fld_pressure_downstream", Float, default=__defaults__["pressure_downstream"]
    )
    pressure_rated = Column(
        "fld_pressure_rated", Float, default=__defaults__["pressure_rated"]
    )
    pressure_upstream = Column(
        "fld_pressure_upstream", Float, default=__defaults__["pressure_upstream"]
    )
    rpm_design = Column("fld_rpm_design", Float, default=__defaults__["rpm_design"])
    rpm_operating = Column(
        "fld_rpm_operating", Float, default=__defaults__["rpm_operating"]
    )
    service_id = Column("fld_service_id", Integer, default=__defaults__["service_id"])
    spring_index = Column(
        "fld_spring_index", Float, default=__defaults__["spring_index"]
    )
    surface_finish = Column(
        "fld_surface_finish", Float, default=__defaults__["surface_finish"]
    )
    technology_id = Column(
        "fld_technology_id", Integer, default=__defaults__["technology_id"]
    )
    thickness = Column("fld_thickness", Float, default=__defaults__["thickness"])
    torque_id = Column("fld_torque_id", Integer, default=__defaults__["torque_id"])
    type_id = Column("fld_type_id", Integer, default=__defaults__["type_id"])
    viscosity_design = Column(
        "fld_viscosity_design", Float, default=__defaults__["viscosity_design"]
    )
    viscosity_dynamic = Column(
        "fld_viscosity_dynamic", Float, default=__defaults__["viscosity_dynamic"]
    )
    water_per_cent = Column(
        "fld_water_per_cent", Float, default=__defaults__["water_per_cent"]
    )
    width_minimum = Column(
        "fld_width_minimum", Float, default=__defaults__["width_minimum"]
    )

    # Define the relationships to other tables in the RAMSTK Program database.

    def get_attributes(self):
        """Retrieve current values of RAMSTKDesignMechanic model attributes.

        :return: {hardware_id, altitude_operating, application_id, balance_id,
                  clearance, casing_id, contact_pressure, deflection,
                  diameter_coil, diameter_inner, diameter_outer, diameter_wire,
                  filter_size, flow_design, flow_operating,
                  frequency_operating, friction, impact_id, leakage_allowable,
                  length, length_compressed, length_relaxed, load_design,
                  load_id, load_operating, lubrication_id, manufacturing_id,
                  material_id, meyer_hardness, misalignment_angle, n_ten,
                  n_cycles, n_elements, offset, particle_size,
                  pressure_contact, pressure_delta, pressure_downstream,
                  pressure_rated, pressure_upstream, rpm_design, rpm_operating,
                  service_id, spring_index, surface_finish, technology_id,
                  thickness, torque_id, type_id, viscosity_design,
                  viscosity_dynamic, water_per_cent, width_minimum} pairs.
        :rtype: dict
        """
        _attributes = {
            "hardware_id": self.hardware_id,
            "altitude_operating": self.altitude_operating,
            "application_id": self.application_id,
            "balance_id": self.balance_id,
            "clearance": self.clearance,
            "casing_id": self.casing_id,
            "contact_pressure": self.contact_pressure,
            "deflection": self.deflection,
            "diameter_coil": self.diameter_coil,
            "diameter_inner": self.diameter_inner,
            "diameter_outer": self.diameter_outer,
            "diameter_wire": self.diameter_wire,
            "filter_size": self.filter_size,
            "flow_design": self.flow_design,
            "flow_operating": self.flow_operating,
            "frequency_operating": self.frequency_operating,
            "friction": self.friction,
            "impact_id": self.impact_id,
            "leakage_allowable": self.leakage_allowable,
            "length": self.length,
            "length_compressed": self.length_compressed,
            "length_relaxed": self.length_relaxed,
            "load_design": self.load_design,
            "load_id": self.load_id,
            "load_operating": self.load_operating,
            "lubrication_id": self.lubrication_id,
            "manufacturing_id": self.manufacturing_id,
            "material_id": self.material_id,
            "meyer_hardness": self.meyer_hardness,
            "misalignment_angle": self.misalignment_angle,
            "n_ten": self.n_ten,
            "n_cycles": self.n_cycles,
            "n_elements": self.n_elements,
            "offset": self.offset,
            "particle_size": self.particle_size,
            "pressure_contact": self.pressure_contact,
            "pressure_delta": self.pressure_delta,
            "pressure_downstream": self.pressure_downstream,
            "pressure_rated": self.pressure_rated,
            "pressure_upstream": self.pressure_upstream,
            "rpm_design": self.rpm_design,
            "rpm_operating": self.rpm_operating,
            "service_id": self.service_id,
            "spring_index": self.spring_index,
            "surface_finish": self.surface_finish,
            "technology_id": self.technology_id,
            "thickness": self.thickness,
            "torque_id": self.torque_id,
            "type_id": self.type_id,
            "viscosity_design": self.viscosity_design,
            "viscosity_dynamic": self.viscosity_dynamic,
            "water_per_cent": self.water_per_cent,
            "width_minimum": self.width_minimum,
        }

        return _attributes



# pylint: disable=duplicate-code
# -*- coding: utf-8 -*-
#
#       ramstk.models.hardware.record.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""RAMSTKHardware Record Model."""

# Standard Library Imports
from datetime import date
from typing import Dict, Union

# Third Party Imports
from sqlalchemy import Column, Float, ForeignKey, Integer, String, UniqueConstraint
from sqlalchemy.orm import relationship

# RAMSTK Package Imports
from ramstk.db import RAMSTK_BASE
from ramstk.models import RAMSTKBaseRecord


class RAMSTKHardwareRecord(RAMSTK_BASE, RAMSTKBaseRecord):
    """Class to represent ramstk_hardware table in the RAMSTK Program database.

    This table shares a:
        * Many-to-One relationship with ramstk_revision.
        * One-to-Many relationship with ramstk_hazard.
        * One-to-Many relationship with ramstk_modes.
        * One-to-Many relationship with ramstk_similar_item.
        * One-to-One relationship with ramstk_allocation.
        * One-to-One relationship with ramstk_reliability.
        * One-to-One relationship with ramstk_mil_hdbk_f.
        * One-to-One relationship with ramstk_nswc.
        * One-to-One relationship with ramstk_design_electric.
        * One-to-One relationship with ramstk_design_mechanic.
    """

    __defaults__ = {
        "alt_part_number": "",
        "attachments": "",
        "cage_code": "",
        "category_id": 0,
        "comp_ref_des": "",
        "cost": 0.0,
        "cost_failure": 0.0,
        "cost_hour": 0.0,
        "cost_type_id": 0,
        "description": "",
        "duty_cycle": 100.0,
        "figure_number": "",
        "lcn": "",
        "level": 0,
        "manufacturer_id": 0,
        "mission_time": 100.0,
        "name": "",
        "nsn": "",
        "page_number": "",
        "parent_id": 0,
        "part": 0,
        "part_number": "",
        "quantity": 1,
        "ref_des": "",
        "remarks": "",
        "repairable": 0,
        "specification_number": "",
        "subcategory_id": 0,
        "tagged_part": 0,
        "total_cost": 0.0,
        "total_part_count": 0,
        "total_power_dissipation": 0.0,
        "year_of_manufacture": date.today().year,
    }
    __tablename__ = "ramstk_hardware"
    __table_args__ = (
        UniqueConstraint(
            "fld_revision_id", "fld_hardware_id", name="ramstk_hardware_ukey"
        ),
        {"extend_existing": True},
    )

    revision_id = Column(
        "fld_revision_id",
        Integer,
        ForeignKey("ramstk_revision.fld_revision_id", ondelete="CASCADE"),
        nullable=False,
    )
    hardware_id = Column(
        "fld_hardware_id", Integer, primary_key=True, autoincrement=True, nullable=False
    )

    alt_part_number = Column(
        "fld_alt_part_number", String(256), default=__defaults__["alt_part_number"]
    )
    attachments = Column(
        "fld_attachments", String(512), default=__defaults__["attachments"]
    )
    cage_code = Column("fld_cage_code", String(256), default=__defaults__["cage_code"])
    category_id = Column(
        "fld_category_id", Integer, default=__defaults__["category_id"]
    )
    comp_ref_des = Column(
        "fld_comp_ref_des", String(256), default=__defaults__["comp_ref_des"]
    )
    cost = Column("fld_cost", Float, default=__defaults__["cost"])
    cost_failure = Column(
        "fld_cost_failure", Float, default=__defaults__["cost_failure"]
    )
    cost_hour = Column("fld_cost_hour", Float, default=__defaults__["cost_hour"])
    cost_type_id = Column(
        "fld_cost_type_id", Integer, default=__defaults__["cost_type_id"]
    )
    description = Column(
        "fld_description", String(512), default=__defaults__["description"]
    )
    duty_cycle = Column("fld_duty_cycle", Float, default=__defaults__["duty_cycle"])
    figure_number = Column(
        "fld_figure_number", String(256), default=__defaults__["figure_number"]
    )
    lcn = Column("fld_lcn", String(256), default=__defaults__["lcn"])
    level = Column("fld_level", Integer, default=__defaults__["level"])
    manufacturer_id = Column(
        "fld_manufacturer_id", Integer, default=__defaults__["manufacturer_id"]
    )
    mission_time = Column(
        "fld_mission_time", Float, default=__defaults__["mission_time"]
    )
    name = Column("fld_name", String(256), default=__defaults__["name"])
    nsn = Column("fld_nsn", String(256), default=__defaults__["nsn"])
    page_number = Column(
        "fld_page_number", String(256), default=__defaults__["page_number"]
    )
    parent_id = Column("fld_parent_id", Integer, default=__defaults__["parent_id"])
    part = Column("fld_part", Integer, default=__defaults__["part"])
    part_number = Column(
        "fld_part_number", String(256), default=__defaults__["part_number"]
    )
    quantity = Column("fld_quantity", Integer, default=__defaults__["quantity"])
    ref_des = Column("fld_ref_des", String(256), default=__defaults__["ref_des"])
    remarks = Column("fld_remarks", String, default=__defaults__["remarks"])
    repairable = Column("fld_repairable", Integer, default=__defaults__["repairable"])
    specification_number = Column(
        "fld_specification_number",
        String(256),
        default=__defaults__["specification_number"],
    )
    subcategory_id = Column(
        "fld_subcategory_id", Integer, default=__defaults__["subcategory_id"]
    )
    tagged_part = Column(
        "fld_tagged_part", Integer, default=__defaults__["tagged_part"]
    )
    total_cost = Column("fld_total_cost", Float, default=__defaults__["total_cost"])
    total_part_count = Column(
        "fld_total_part_count", Integer, default=__defaults__["total_part_count"]
    )
    total_power_dissipation = Column(
        "fld_total_power_dissipation",
        Float,
        default=__defaults__["total_power_dissipation"],
    )
    year_of_manufacture = Column(
        "fld_year_of_manufacture", Integer, default=__defaults__["year_of_manufacture"]
    )

    # Define the relationships to other tables in the RAMSTK Program database.
    allocation: relationship = relationship(
        "RAMSTKAllocationRecord",
        uselist=False,
        backref="hardware",
        passive_deletes=True,
    )
    sia: relationship = relationship(
        "RAMSTKSimilarItemRecord",
        uselist=False,
        backref="hardware",
        passive_deletes=True,
    )
    reliability: relationship = relationship(
        "RAMSTKReliabilityRecord",
        uselist=False,
        backref="hardware",
        passive_deletes=True,
    )
    milhdbkf: relationship = relationship(
        "RAMSTKMilHdbk217FRecord",
        uselist=False,
        backref="hardware",
        passive_deletes=True,
    )
    nswc: relationship = relationship(
        "RAMSTKNSWCRecord",
        uselist=False,
        backref="hardware",
        passive_deletes=True,
    )
    design_electric: relationship = relationship(
        "RAMSTKDesignElectricRecord",
        uselist=False,
        backref="hardware",
        passive_deletes=True,
    )
    design_mechanic: relationship = relationship(
        "RAMSTKDesignMechanicRecord",
        uselist=False,
        backref="hardware",
        passive_deletes=True,
    )

    def get_attributes(self) -> Dict[str, Union[float, int, str]]:
        """Retrieve the current values of RAMSTKHardware data model attributes.

        :return: {revision_id, hardware_id, alt_part_number, attachments,
                  cage_code, category_id, comp_ref_des, cost, cost_failure,
                  cost_hour, cost_type_id, description, duty_cycle,
                  figure_number, lcn, level, manufacturer_id, mission_time,
                  name, nsn, page_number, parent_id, part, part_number,
                  quantity, ref_des, remarks, repairable, specification_number,
                  subcategory_id, tagged_part, total_cost, total_part_count,
                  total_power_dissipation, year_of_manufacture} pairs.
        :rtype: dict
        """
        _attributes = {
            "revision_id": self.revision_id,
            "hardware_id": self.hardware_id,
            "alt_part_number": self.alt_part_number,
            "attachments": self.attachments,
            "cage_code": self.cage_code,
            "category_id": self.category_id,
            "comp_ref_des": self.comp_ref_des,
            "cost": self.cost,
            "cost_failure": self.cost_failure,
            "cost_hour": self.cost_hour,
            "cost_type_id": self.cost_type_id,
            "description": self.description,
            "duty_cycle": self.duty_cycle,
            "figure_number": self.figure_number,
            "lcn": self.lcn,
            "level": self.level,
            "manufacturer_id": self.manufacturer_id,
            "mission_time": self.mission_time,
            "name": self.name,
            "nsn": self.nsn,
            "page_number": self.page_number,
            "parent_id": self.parent_id,
            "part": self.part,
            "part_number": self.part_number,
            "quantity": self.quantity,
            "ref_des": self.ref_des,
            "remarks": self.remarks,
            "repairable": self.repairable,
            "specification_number": self.specification_number,
            "subcategory_id": self.subcategory_id,
            "tagged_part": self.tagged_part,
            "total_cost": self.total_cost,
            "total_part_count": self.total_part_count,
            "total_power_dissipation": self.total_power_dissipation,
            "year_of_manufacture": self.year_of_manufacture,
        }

        return _attributes

    def do_calculate_total_cost(self) -> None:
        """Calculate the total cost of the hardware item."""
        if self.cost_type_id == 1:
            self.total_cost = self.cost
        elif self.cost_type_id == 2:
            self.total_cost = self.cost * self.quantity



# Standard Library Imports
from typing import Any, List

# RAMSTK Package Imports
from ramstk.configuration import RAMSTKUserConfiguration as RAMSTKUserConfiguration
from ramstk.views.gtk3 import GObject as GObject
from ramstk.views.gtk3 import Gtk as Gtk
from ramstk.views.gtk3 import _ as _
from ramstk.views.gtk3.widgets import RAMSTKComboBox as RAMSTKComboBox
from ramstk.views.gtk3.widgets import RAMSTKLabel as RAMSTKLabel

class ImportProject(Gtk.Assistant):
    RAMSTK_USER_CONFIGURATION: Any = ...
    cmbSelectModule: Any = ...
    _filechooser: Any = ...
    _module: str = ...
    tvwFieldMap: Any = ...
    def __init__(
        self,
        __button: Gtk.ToolButton,
        configuration: RAMSTKUserConfiguration,
        parent: object,
    ) -> None: ...
    def _do_edit_cell(
        self, __cell: Gtk.CellRenderer, path: str, new_text: str, model: Gtk.TreeModel
    ) -> None: ...
    def _do_load_db_fields(self, db_fields: List[str]) -> None: ...
    def _do_load_import_fields(self, import_fields: List[str]) -> None: ...
    def _do_quit(self, __widget: Gtk.Widget) -> None: ...
    def _do_request_import(self, __assistant: Gtk.Assistant) -> None: ...
    def _do_select_file(self, filechooser: Gtk.FileChooser) -> None: ...
    def _on_combo_changed(self, combo: RAMSTKComboBox) -> None: ...
    def __do_load_combobox(self) -> None: ...
    def __do_set_callbacks(self) -> None: ...
    def __do_set_properties(self) -> None: ...
    def __make_confirm_page(self) -> None: ...
    def __make_input_file_select_page(self) -> None: ...
    def __make_introduction_page(self) -> None: ...
    def __make_map_field_page(self) -> None: ...
    def __make_ui(self) -> None: ...



# pylint: disable=duplicate-code
# -*- coding: utf-8 -*-
#
#       ramstk.models.action.record.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""RAMSTKAction Table Module."""

# Standard Library Imports
from datetime import date, timedelta

# Third Party Imports
from sqlalchemy import Column, Date, ForeignKeyConstraint, Integer, String
from sqlalchemy.orm import relationship

# RAMSTK Package Imports
from ramstk.db import RAMSTK_BASE
from ramstk.models import RAMSTKBaseRecord


class RAMSTKActionRecord(RAMSTK_BASE, RAMSTKBaseRecord):
    """Class to represent table ramstk_action in the RAMSTK Program database.

    This table shares a Many-to-One relationship with ramstk_cause.
    """

    __defaults__ = {
        "action_recommended": "",
        "action_category": "",
        "action_owner": "",
        "action_due_date": date.today() + timedelta(days=30),
        "action_status": "",
        "action_taken": "",
        "action_approved": 0,
        "action_approve_date": date.today() + timedelta(days=30),
        "action_closed": 0,
        "action_close_date": date.today() + timedelta(days=30),
    }
    __tablename__ = "ramstk_action"
    __table_args__ = (
        ForeignKeyConstraint(
            [
                "fld_revision_id",
                "fld_hardware_id",
                "fld_mode_id",
                "fld_mechanism_id",
                "fld_cause_id",
            ],
            [
                "ramstk_cause.fld_revision_id",
                "ramstk_cause.fld_hardware_id",
                "ramstk_cause.fld_mode_id",
                "ramstk_cause.fld_mechanism_id",
                "ramstk_cause.fld_cause_id",
            ],
        ),
        {"extend_existing": True},
    )

    revision_id = Column("fld_revision_id", Integer, primary_key=True, nullable=False)
    hardware_id = Column(
        "fld_hardware_id", Integer, primary_key=True, default=-1, nullable=False
    )
    mode_id = Column("fld_mode_id", Integer, primary_key=True, nullable=False)
    mechanism_id = Column("fld_mechanism_id", Integer, primary_key=True, nullable=False)
    cause_id = Column(
        "fld_cause_id", Integer, primary_key=True, nullable=False, unique=True
    )
    action_id = Column(
        "fld_action_id", Integer, primary_key=True, autoincrement=True, nullable=False
    )

    action_recommended = Column(
        "fld_action_recommended", String, default=__defaults__["action_recommended"]
    )
    action_category = Column(
        "fld_action_category", String(512), default=__defaults__["action_category"]
    )
    action_owner = Column(
        "fld_action_owner", String(512), default=__defaults__["action_owner"]
    )
    action_due_date = Column(
        "fld_action_due_date", Date, default=__defaults__["action_due_date"]
    )
    action_status = Column(
        "fld_action_status", String(512), default=__defaults__["action_status"]
    )
    action_taken = Column(
        "fld_action_taken", String, default=__defaults__["action_taken"]
    )
    action_approved = Column(
        "fld_action_approved", Integer, default=__defaults__["action_approved"]
    )
    action_approve_date = Column(
        "fld_action_approve_date", Date, default=__defaults__["action_approve_date"]
    )
    action_closed = Column(
        "fld_action_closed", Integer, default=__defaults__["action_closed"]
    )
    action_close_date = Column(
        "fld_action_close_date", Date, default=__defaults__["action_close_date"]
    )

    # Define the relationships to other tables in the RAMSTK Program database.
    cause = relationship("RAMSTKCauseRecord", back_populates="action")  # type: ignore

    is_mode = False
    is_mechanism = False
    is_cause = False
    is_control = False
    is_action = True

    def get_attributes(self):
        """Retrieve current values of the RAMSTKAction data model attributes.

        :return: {cause_id, action_id, action_recommended,
                  action_category, action_owner, action_due_date,
                  action_status, action_taken, action_approved,
                  action_approved_date, action_closed,
                  action_closed_date} pairs.
        :rtype: dict
        """
        _attributes = {
            "cause_id": self.cause_id,
            "action_id": self.action_id,
            "action_recommended": self.action_recommended,
            "action_category": self.action_category,
            "action_owner": self.action_owner,
            "action_due_date": self.action_due_date,
            "action_status": self.action_status,
            "action_taken": self.action_taken,
            "action_approved": self.action_approved,
            "action_approve_date": self.action_approve_date,
            "action_closed": self.action_closed,
            "action_close_date": self.action_close_date,
        }

        return _attributes



# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.design_electric.components.integrated_circuit.py is part of
#       the RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Integrated Circuit Input Panel."""

# Standard Library Imports
from typing import Any, Dict, List

# Third Party Imports
from pubsub import pub

# RAMSTK Package Imports
from ramstk.views.gtk3 import _
from ramstk.views.gtk3.widgets import RAMSTKComboBox, RAMSTKEntry, RAMSTKFixedPanel


class ICDesignElectricInputPanel(RAMSTKFixedPanel):
    """Display IC assessment input attribute data in the RAMSTK Work Book.

    The Integrated Circuit assessment input view displays all the assessment
    inputs for the selected integrated circuit.  This includes, currently,
    inputs for MIL-HDBK-217FN2.  The attributes of an integrated circuit
    assessment input view are:

    :cvar dict _dic_technology: dictionary of integrated circuit package
        technologies.  Key is integrated circuit subcategory ID; values are
        lists of technologies.

    :ivar cmbApplication: select and display the application of the integrated
        circuit.
    :ivar cmbConstruction: select and display the construction of the
        integrated circuit.
    :ivar cmbECC: select and display the error correction code used by the
        EEPROM.
    :ivar cmbManufacturing: select and display the manufacturing approach for
        the integrated circuit.
    :ivar cmbPackage: select and display the package type of the integrated
        circuit.
    :ivar cmbTechnology: select and display the technology used in the
        integrated circuit.
    :ivar cmbType: select and display the type of the integrated circuit.

    :ivar txtArea: enter and display the die area of the integrated circuit.
    :ivar txtFeatureSize: enter and display the feature size (in microns) of
        the VLSI.
    :ivar txtNActivePins: enter and display the number of active pins.
    :ivar txtNCycles: enter and display the number of programming cycles over
        the life of the PROM.
    :ivar txtNElements: enter and display the number of elements (transistors,
        gates, etc.) in the integrated circuit.
    :ivar txtOperatingLife: enter and display the operating life of the
        integrated circuit.
    :ivar txtThetaJC: enter and display the junction - case thermal resistance
        of the integrated circuit.
    :ivar txtVoltageESD: enter and display the ESD threshold voltage of the
        VLSI.
    :ivar txtYearsInProduction: enter and display the number of years the
        integrated circuit type has been in production.
    """

    # Define private dict class attributes.
    _dic_technology: Dict[int, List[List[str]]] = {
        1: [["MOS"], [_("Bipolar")]],
        2: [
            ["TTL"],
            ["ASTTL"],
            ["CML"],
            ["HTTL"],
            ["FTTL"],
            ["DTL"],
            ["ECL"],
            ["ALSTTL"],
            ["FLTTL"],
            ["STTL"],
            ["BiCMOS"],
            ["LSTTL"],
            ["III"],
            ["IIIL"],
            ["ISL"],
        ],
        3: [["MOS"], [_("Bipolar")]],
        4: [["MOS"], [_("Bipolar")]],
        5: [["MOS"], [_("Bipolar")]],
        6: [["MOS"], [_("Bipolar")]],
        7: [["MOS"], [_("Bipolar")]],
        8: [["MOS"], [_("Bipolar")]],
        9: [["MMIC"], [_("Digital")]],
    }
    _dic_types: Dict[int, List[List[str]]] = {
        9: [["MMIC"], [_("Digital")]],
        10: [[_("Logic and Custom")], [_("Gate Array")]],
    }

    # Define private list class attributes.

    # Define private scalar class attributes.
    _record_field = "hardware_id"
    _select_msg = "selected_hardware"
    _tag = "design_electric"
    _title = _("Integrated Circuit Design Inputs")

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self) -> None:
        """Initialize an instance of the IC assessment input view."""
        super().__init__()

        # Initialize widgets.
        self.cmbApplication: RAMSTKComboBox = RAMSTKComboBox()
        self.cmbConstruction: RAMSTKComboBox = RAMSTKComboBox()
        self.cmbECC: RAMSTKComboBox = RAMSTKComboBox()
        self.cmbManufacturing: RAMSTKComboBox = RAMSTKComboBox()
        self.cmbPackage: RAMSTKComboBox = RAMSTKComboBox()
        self.cmbQuality: RAMSTKComboBox = RAMSTKComboBox()
        self.cmbTechnology: RAMSTKComboBox = RAMSTKComboBox()
        self.cmbType: RAMSTKComboBox = RAMSTKComboBox()
        self.txtArea: RAMSTKEntry = RAMSTKEntry()
        self.txtFeatureSize: RAMSTKEntry = RAMSTKEntry()
        self.txtNActivePins: RAMSTKEntry = RAMSTKEntry()
        self.txtNCycles: RAMSTKEntry = RAMSTKEntry()
        self.txtNElements: RAMSTKEntry = RAMSTKEntry()
        self.txtOperatingLife: RAMSTKEntry = RAMSTKEntry()
        self.txtThetaJC: RAMSTKEntry = RAMSTKEntry()
        self.txtVoltageESD: RAMSTKEntry = RAMSTKEntry()
        self.txtYearsInProduction: RAMSTKEntry = RAMSTKEntry()

        # Initialize private dictionary attributes.

        # Initialize private list attributes.

        # Initialize private scalar attributes.
        self._hazard_rate_method_id: int = 0
        self._quality_id: int = 0

        # Initialize public dictionary attributes.
        self.dic_attribute_widget_map: Dict[str, List[Any]] = {
            "quality_id": [
                32,
                self.cmbQuality,
                "changed",
                super().on_changed_combo,
                "wvw_editing_reliability",
                0,
                {
                    "tooltip": _("The quality level of the integrated circuit."),
                },
                _("Quality Level:"),
                "gint",
            ],
            "package_id": [
                30,
                self.cmbPackage,
                "changed",
                super().on_changed_combo,
                f"wvw_editing_{self._tag}",
                0,
                {
                    "tooltip": _("The type of package housing the integrated circuit."),
                },
                _("Package:"),
                "gint",
            ],
            "area": [
                3,
                self.txtArea,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                0.0,
                {
                    "tooltip": _(
                        "The die area (in mil<sup>2</sup>) of the integrated circuit."
                    ),
                    "width": 125,
                },
                _("Die Area:"),
                "gfloat",
            ],
            "n_elements": [
                25,
                self.txtNElements,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                0,
                {
                    "tooltip": _(
                        "The number of active elements in the integrated circuit."
                    ),
                    "width": 125,
                },
                _("N Elements:"),
                "gint",
            ],
            "theta_jc": [
                47,
                self.txtThetaJC,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                0.0,
                {
                    "tooltip": _("The junction to case thermal resistance."),
                    "width": 125,
                },
                _("\u0398<sub>JC</sub>:"),
                "gfloat",
            ],
            "n_active_pins": [
                22,
                self.txtNActivePins,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                0,
                {
                    "tooltip": _(
                        "The number of active pins on the integrated circuit."
                    ),
                    "width": 125,
                },
                _("Active Pins:"),
                "gint",
            ],
            "technology_id": [
                37,
                self.cmbTechnology,
                "changed",
                super().on_changed_combo,
                f"wvw_editing_{self._tag}",
                0,
                {
                    "tooltip": _(
                        "The technology used to construct the integrated circuit."
                    ),
                },
                _("Technology:"),
                "gint",
            ],
            "years_in_production": [
                55,
                self.txtYearsInProduction,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                2,
                {
                    "tooltip": _(
                        "The number of years the generic device type has been in "
                        "production."
                    ),
                    "width": 125,
                },
                _("Years in Production:"),
                "gfloat",
            ],
            "construction_id": [
                6,
                self.cmbConstruction,
                "changed",
                super().on_changed_combo,
                f"wvw_editing_{self._tag}",
                0,
                {
                    "tooltip": _(
                        "The method of construction of the integrated circuit."
                    ),
                },
                _("Construction:"),
                "gint",
            ],
            "n_cycles": [
                24,
                self.txtNCycles,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                0,
                {
                    "tooltip": _(
                        "The total number of programming cycles over the EEPROM life."
                    ),
                    "width": 125,
                },
                _("Programming Cycles:"),
                "gint",
            ],
            "operating_life": [
                28,
                self.txtOperatingLife,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                0.0,
                {
                    "tooltip": _("The system lifetime operating hours."),
                    "width": 125,
                },
                _("Operating Life:"),
                "gfloat",
            ],
            "family_id": [
                15,
                self.cmbECC,
                "changed",
                super().on_changed_combo,
                f"wvw_editing_{self._tag}",
                0,
                {
                    "tooltip": _("The error correction code used by the EEPROM."),
                },
                _("Error Correction Code:"),
                "gint",
            ],
            "application_id": [
                2,
                self.cmbApplication,
                "changed",
                super().on_changed_combo,
                f"wvw_editing_{self._tag}",
                0,
                {
                    "tooltip": _("The application of the integrated circuit."),
                },
                _("Application:"),
                "gint",
            ],
            "type_id": [
                48,
                self.cmbType,
                "changed",
                super().on_changed_combo,
                f"wvw_editing_{self._tag}",
                0,
                {
                    "tooltip": _("The type of GaAs or VLSI device."),
                },
                _("Device Type:"),
                "gint",
            ],
            "feature_size": [
                16,
                self.txtFeatureSize,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                0.0,
                {
                    "tooltip": _("The feature size (in microns) of the VLSI device."),
                    "width": 125,
                },
                _("Feature Size:"),
                "gfloat",
            ],
            "manufacturing_id": [
                20,
                self.cmbManufacturing,
                "changed",
                super().on_changed_combo,
                f"wvw_editing_{self._tag}",
                0,
                {
                    "tooltip": _("The manufacturing process for the VLSI device."),
                },
                _("Manufacturing Process:"),
                "gint",
            ],
            "voltage_esd": [
                51,
                self.txtVoltageESD,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                0.0,
                {
                    "tooltip": _(
                        "The ESD susceptibility threshold voltage of the VLSI device."
                    ),
                    "width": 125,
                },
                _("ESD Threshold Voltage:"),
                "gfloat",
            ],
        }

        # Initialize public list attributes.

        # Initialize public scalar attributes.
        self.category_id: int = 0
        self.subcategory_id: int = 0

        super().do_set_properties()
        super().do_make_panel()
        super().do_set_callbacks()

        # Subscribe to PyPubSub messages.
        pub.subscribe(
            self.do_load_comboboxes,
            "changed_subcategory",
        )
        pub.subscribe(
            self._do_set_reliability_attributes,
            "succeed_get_reliability_attributes",
        )

    # pylint: disable=unused-argument
    def do_load_comboboxes(self, subcategory_id: int) -> None:
        """Load the integrated circuit RAMSTKComboBox()s.

        :param subcategory_id: the subcategory ID of the selected IC.  This is
            unused in this method but required because this method is a
            PyPubSub listener.
        :return: None
        :rtype: None
        """
        self.subcategory_id = subcategory_id

        # Load the quality level RAMSTKComboBox().
        self.cmbQuality.do_load_combo(
            [[_("Class S")], [_("Class B")], [_("Class B-1")]]
        )

        # Load the Construction RAMSTKComboBox().
        self.cmbConstruction.do_load_combo(
            [["FLOTOX"], [_("Textured Poly")]], signal="changed"
        )

        # Load the error correction code RAMSTKComboBox().
        self.cmbECC.do_load_combo(
            [
                [_("No on-chip ECC")],
                [_("On-chip Hamming code")],
                [_("Two-Needs-One redundant cell approach")],
            ],
            signal="changed",
        )

        # Load the manufacturing process RAMSTKComboBox().
        self.cmbManufacturing.do_load_combo(
            [["QML or QPL"], ["Non-QML or non-QPL"]], signal="changed"
        )

        # Load the package RAMSTKComboBox().
        self.cmbPackage.do_load_combo(
            [
                [_("Hermetic DIP w/ Solder or Weld Seal")],
                [_("Hermetic Pin Grid Array (PGA)")],
                [_("Hermetic SMT (Leaded and Nonleaded)")],
                [_("DIP w/ Glass Seal")],
                [_("Flatpacks w/ Axial Leads")],
                ["Can"],
                [_("Nonhermetic DIP")],
                [_("Nonhermetic Pin Grid Array (PGA)")],
                [_("Nonhermetic SMT")],
            ],
            signal="changed",
        )

        # Load the technology RAMSTKComboBox().
        try:
            if self._hazard_rate_method_id == 1:
                if subcategory_id == 9:
                    _data = [["MMIC"], [_("Digital")]]
                else:
                    _data = [["Bipolar"], ["MOS"]]
            else:
                _data = self._dic_technology[subcategory_id]
        except KeyError:
            _data = []
        self.cmbTechnology.do_load_combo(_data, signal="changed")

        # Load the device type RAMSTKComboBox().
        try:
            _data = self._dic_types[subcategory_id]
        except KeyError:
            _data = []
        self.cmbType.do_load_combo(_data, signal="changed")

    def _do_set_reliability_attributes(self, attributes: Dict[str, Any]) -> None:
        """Set the attributes when the reliability attributes are retrieved.

        :param attributes: the dict of reliability attributes.
        :return: None
        :rtype: None
        """
        if attributes["hardware_id"] == self._record_id:
            self._hazard_rate_method_id = attributes["hazard_rate_method_id"]
            self._quality_id = attributes["quality_id"]

            self.cmbQuality.set_sensitive(True)
            self.cmbQuality.do_update(
                self._quality_id,
                signal="changed",
            )

            pub.sendMessage(
                f"request_get_{self._tag}_attributes",
                node_id=self._record_id,
            )

    def _do_set_sensitive(self, attributes: Dict[str, Any]) -> None:
        """Set widget sensitivity as needed for the selected IC.

        :return: None
        :rtype: None
        """
        self.cmbApplication.set_sensitive(False)
        self.cmbConstruction.set_sensitive(False)
        self.cmbECC.set_sensitive(False)
        self.cmbManufacturing.set_sensitive(False)
        self.cmbPackage.set_sensitive(False)
        self.cmbTechnology.set_sensitive(False)
        self.cmbType.set_sensitive(False)
        self.txtArea.set_sensitive(False)
        self.txtFeatureSize.set_sensitive(False)
        self.txtNActivePins.set_sensitive(False)
        self.txtNCycles.set_sensitive(False)
        self.txtNElements.set_sensitive(False)
        self.txtOperatingLife.set_sensitive(False)
        self.txtThetaJC.set_sensitive(False)
        self.txtVoltageESD.set_sensitive(False)
        self.txtYearsInProduction.set_sensitive(False)

        _dic_method = {
            1: self.__do_load_linear,
            2: self.__do_load_logic,
            3: self.__do_load_pal_pla,
            4: self.__do_load_microprocessor_microcontroller,
            5: self.__do_load_rom,
            6: self.__do_load_eeprom,
            7: self.__do_load_dram,
            8: self.__do_load_sram,
            9: self.__do_load_gaas,
            10: self.__do_load_vhsic_vlsi,
        }
        try:
            # noinspection PyArgumentList
            _dic_method[self.subcategory_id](attributes)
        except KeyError:
            pass

        if self._hazard_rate_method_id == 2:  # MIL-HDBK-217F, Part Stress
            self.cmbPackage.set_sensitive(True)
            self.cmbPackage.do_update(
                attributes["package_id"],
                signal="changed",
            )
            self.txtNElements.set_sensitive(True)
            self.txtNElements.do_update(
                str(attributes["n_elements"]),
                signal="changed",
            )
            self.txtThetaJC.set_sensitive(True)
            self.txtThetaJC.do_update(
                str(attributes["theta_jc"]),
                signal="changed",
            )
            self.txtNActivePins.do_update(
                str(attributes["n_active_pins"]),
                signal="changed",
            )

    def _do_load_application_combo(self, attributes: Dict[str, Any]) -> None:
        """Load the IC application RAMSTKComboBox().

        :param attributes: the attributes dict for the selected integrated circuit.
        :return: None
        """
        if attributes["construction_id"] == 1:
            self.cmbApplication.do_load_combo(
                [
                    [_("Low Noise and Low Power (\u2264 100mW)")],
                    [_("Driver and High Power (> 100mW)")],
                    [_("Unknown")],
                ],
                signal="changed",
            )
        else:
            self.cmbApplication.do_load_combo(
                [[_("All digital devices")]],
                signal="changed",
            )

    def __do_load_dram(self, attributes: Dict[str, Any]) -> None:
        """Load the widgets that display DRAM information.

        :param attributes: the attributes dictionary for the selected
            integrated circuit.
        :return: None
        """
        if self._hazard_rate_method_id == 2:  # MIL-HDBK-217F, Part Stress
            self.cmbTechnology.do_update(
                attributes["technology_id"],
                signal="changed",
            )
        self.__do_set_dram_sensitive()

    def __do_load_eeprom(self, attributes: Dict[str, Any]) -> None:
        """Load the widgets that display EEPROM information.

        :param attributes: the attributes dictionary for the selected
            Integrated Circuit.
        :return: None
        :rtype: None
        """
        if self._hazard_rate_method_id == 2:  # MIL-HDBK-217F, Part Stress
            self.cmbConstruction.do_update(
                attributes["construction_id"],
                signal="changed",
            )
            self.cmbTechnology.do_update(
                attributes["technology_id"],
                signal="changed",
            )
            self.cmbECC.do_update(
                attributes["family_id"],
                signal="changed",
            )
            self.txtNCycles.do_update(
                str(attributes["n_cycles"]),
                signal="changed",
            )
            self.txtOperatingLife.do_update(
                str(self.fmt.format(attributes["operating_life"])),
                signal="changed",
            )
        self.__do_set_eeprom_sensitive()

    def __do_load_gaas(self, attributes: Dict[str, Any]) -> None:
        """Load the widgets that display GaAs IC information.

        :param attributes: the attributes dictionary for the selected
            Integrated Circuit.
        :return: None
        :rtype: None
        """
        if self._hazard_rate_method_id == 2:  # MIL-HDBK-217F, Part Stress
            self.cmbApplication.do_update(
                attributes["application_id"],
                signal="changed",
            )
            self.cmbType.do_update(
                attributes["type_id"],
                signal="changed",
            )
            self.txtYearsInProduction.do_update(
                str(attributes["years_in_production"]),
                signal="changed",
            )
        self.__do_set_gaas_sensitive()

    def __do_load_linear(self, attributes: Dict[str, Any]) -> None:
        """Load the widgets that display linear (analog) information.

        :param attributes: the attributes dictionary for the selected
            Integrated Circuit.
        :return: None
        :rtype: None
        """
        self.cmbTechnology.do_update(
            attributes["technology_id"],
            signal="changed",
        )
        if self._hazard_rate_method_id == 2:  # MIL-HDBK-217F, Part Stress
            self.txtNActivePins.do_update(
                str(attributes["n_active_pins"]),
                signal="changed",
            )
            self.txtYearsInProduction.do_update(
                str(attributes["years_in_production"]),
                signal="changed",
            )

        self.__do_set_linear_sensitive()

    def __do_load_logic(self, attributes: Dict[str, Any]) -> None:
        """Load the widgets that display digital IC information.

        :param attributes: the attributes dictionary for the selected
            Integrated Circuit.
        :return: None
        :rtype: None
        """
        self.cmbTechnology.do_update(
            attributes["technology_id"],
            signal="changed",
        )
        if self._hazard_rate_method_id == 2:  # MIL-HDBK-217F, Part Stress
            self.txtNActivePins.do_update(
                str(attributes["n_active_pins"]),
                signal="changed",
            )
            self.txtYearsInProduction.do_update(
                str(attributes["years_in_production"]),
                signal="changed",
            )

        self.__do_set_logic_sensitive()

    def __do_load_microprocessor_microcontroller(
        self, attributes: Dict[str, Any]
    ) -> None:
        """Load the widgets that display microprocessor information.

        :param attributes: the attributes dictionary for the selected
            Integrated Circuit.
        :return: None
        :rtype: None
        """
        self.cmbTechnology.do_update(
            attributes["technology_id"],
            signal="changed",
        )
        if self._hazard_rate_method_id == 2:  # MIL-HDBK-217F, Part Stress
            self.txtNActivePins.do_update(
                str(attributes["n_active_pins"]),
                signal="changed",
            )
            self.txtYearsInProduction.do_update(
                str(attributes["years_in_production"]),
                signal="changed",
            )

        self.__do_set_microprocessor_microcontroller_sensitive()

    def __do_load_pal_pla(self, attributes: Dict[str, Any]) -> None:
        """Load the widgets that display DRAM information.

        :param attributes: the attributes dictionary for the selected
            Integrated Circuit.
        :return: None
        :rtype: None
        """
        self.cmbTechnology.do_update(
            attributes["technology_id"],
            signal="changed",
        )
        if self._hazard_rate_method_id == 2:  # MIL-HDBK-217F, Part Stress
            self.txtNActivePins.do_update(
                str(attributes["n_active_pins"]),
                signal="changed",
            )
            self.txtYearsInProduction.do_update(
                str(attributes["years_in_production"]),
                signal="changed",
            )

        self.__do_set_pal_pla_sensitive()

    def __do_load_rom(self, attributes: Dict[str, Any]) -> None:
        """Load the widgets that display ROM information.

        :param attributes: the attributes dictionary for the selected
            Integrated Circuit.
        :return: None
        :rtype: None
        """
        self.cmbTechnology.do_update(
            attributes["technology_id"],
            signal="changed",
        )
        self.__do_set_rom_sensitive()

    def __do_load_sram(self, attributes: Dict[str, Any]) -> None:
        """Load the widgets that display SRAM information.

        :param attributes: the attributes dictionary for the selected
            Integrated Circuit.
        :return: None
        :rtype: None
        """
        self.cmbTechnology.do_update(
            attributes["technology_id"],
            signal="changed",
        )
        self.__do_set_sram_sensitive()

    def __do_load_vhsic_vlsi(self, attributes: Dict[str, Any]) -> None:
        """Load the widgets that display VHSIC/VLSI information.

        :param attributes: the attributes dictionary for the selected
            Integrated Circuit.
        :return: None
        :rtype: None
        """
        self.cmbManufacturing.do_update(
            attributes["manufacturing_id"],
            signal="changed",
        )
        self.cmbType.do_update(attributes["type_id"])
        self.txtArea.do_update(
            str(self.fmt.format(attributes["area"])),
            signal="changed",
        )
        self.txtFeatureSize.do_update(
            str(self.fmt.format(attributes["feature_size"])),
            signal="changed",
        )
        self.txtVoltageESD.do_update(
            str(self.fmt.format(attributes["voltage_esd"])),
            signal="changed",
        )
        self.__do_set_vhsic_vlsi_sensitive()

    def __do_set_dram_sensitive(self) -> None:
        """Set the widgets that display DRAM information sensitive.

        :return: None
        :rtype: None
        """
        if self._hazard_rate_method_id == 1:  # MIL-HDBK-217F, Parts Count
            pub.sendMessage(
                "wvw_editing_hardware",
                node_id=self._record_id,
                package={"technology_id": 2},
            )
            self.txtNElements.set_sensitive(True)
        if self._hazard_rate_method_id == 2:  # MIL-HDBK-217F, Part Stress
            self.cmbTechnology.set_sensitive(True)

    def __do_set_eeprom_sensitive(self) -> None:
        """Set the widgets that display EEPROM information sensitive.

        :return: None
        :rtype: None
        """
        if self._hazard_rate_method_id == 1:  # MIL-HDBK-217F, Parts Count
            pub.sendMessage(
                "wvw_editing_design_electric",
                node_id=self._record_id,
                package={"technology_id": 2},
            )
            self.txtNElements.set_sensitive(True)
        elif self._hazard_rate_method_id == 2:  # MIL-HDBK-217F, Part Stress
            self.cmbConstruction.set_sensitive(True)
            self.cmbECC.set_sensitive(True)
            self.cmbTechnology.set_sensitive(True)
            self.txtNActivePins.set_sensitive(True)
            self.txtNCycles.set_sensitive(True)
            self.txtOperatingLife.set_sensitive(True)

    def __do_set_gaas_sensitive(self) -> None:
        """Set the widgets that display GaAs IC information sensitive.

        :return: None
        :rtype: None
        """
        if self._hazard_rate_method_id == 1:  # MIL-HDBK-217F, Parts Count
            self.cmbTechnology.set_sensitive(True)
            self.txtNElements.set_sensitive(True)
        elif self._hazard_rate_method_id == 2:  # MIL-HDBK-217F, Part Stress
            self.cmbApplication.set_sensitive(True)
            self.cmbType.set_sensitive(True)
            self.txtNActivePins.set_sensitive(True)
            self.txtYearsInProduction.set_sensitive(True)

    def __do_set_linear_sensitive(self) -> None:
        """Set the widgets that display linear (analog) information sensitive.

        :return: None
        :rtype: None
        """
        self.cmbTechnology.set_sensitive(True)
        if self._hazard_rate_method_id == 1:  # MIL-HDBK-217F, Parts Count
            self.txtNElements.set_sensitive(True)

        elif self._hazard_rate_method_id == 2:  # MIL-HDBK-217F, Part Stress
            self.txtNActivePins.set_sensitive(True)
            self.txtYearsInProduction.set_sensitive(True)

    def __do_set_logic_sensitive(self) -> None:
        """Set the widgets that display digital IC information sensitive.

        :return: None
        :rtype: None
        """
        self.cmbTechnology.set_sensitive(True)
        if self._hazard_rate_method_id == 1:  # MIL-HDBK-217F, Parts Count
            self.txtNElements.set_sensitive(True)
        elif self._hazard_rate_method_id == 2:  # MIL-HDBK-217F, Part Stress
            self.txtNActivePins.set_sensitive(True)
            self.txtYearsInProduction.set_sensitive(True)

    def __do_set_microprocessor_microcontroller_sensitive(self) -> None:
        """Set the widgets that display microprocessor information sensitive.

        :return: None
        :rtype: None
        """
        self.cmbTechnology.set_sensitive(True)
        if self._hazard_rate_method_id == 1:  # MIL-HDBK-217F, Parts Count
            self.txtNElements.set_sensitive(True)
        elif self._hazard_rate_method_id == 2:  # MIL-HDBK-217F, Part Stress
            self.txtNActivePins.set_sensitive(True)
            self.txtYearsInProduction.set_sensitive(True)

    def __do_set_pal_pla_sensitive(self) -> None:
        """Set the widgets that display DRAM information sensitive.

        :return: None
        :rtype: None
        """
        self.cmbTechnology.set_sensitive(True)
        if self._hazard_rate_method_id == 1:  # MIL-HDBK-217F, Parts Count
            self.txtNElements.set_sensitive(True)
        elif self._hazard_rate_method_id == 2:  # MIL-HDBK-217F, Part Stress
            self.txtNActivePins.set_sensitive(True)
            self.txtYearsInProduction.set_sensitive(True)

    def __do_set_rom_sensitive(self) -> None:
        """Set the widgets that display ROM information sensitive.

        :return: None
        :rtype: None
        """
        self.cmbTechnology.set_sensitive(True)
        if self._hazard_rate_method_id == 1:  # MIL-HDBK-217F, Parts Count
            self.txtNElements.set_sensitive(True)

    def __do_set_sram_sensitive(self) -> None:
        """Set the widgets that display SRAM information sensitive.

        :return: None
        :rtype: None
        """
        self.cmbTechnology.set_sensitive(True)
        if self._hazard_rate_method_id == 1:  # MIL-HDBK-217F, Parts Count
            self.txtNElements.set_sensitive(True)

    def __do_set_vhsic_vlsi_sensitive(self) -> None:
        """Set the widgets that display VHSIC/VLSI information sensitive.

        :return: None
        :rtype: None
        """
        self.cmbManufacturing.set_sensitive(True)
        self.cmbType.set_sensitive(True)
        self.txtArea.set_sensitive(True)
        self.txtFeatureSize.set_sensitive(True)
        self.txtNActivePins.set_sensitive(False)
        self.txtVoltageESD.set_sensitive(True)



# Standard Library Imports
from typing import Any, Callable, Dict, List, Tuple

# Third Party Imports
import treelib

# RAMSTK Package Imports
from ramstk.views.gtk3 import Gtk as Gtk
from ramstk.views.gtk3 import _ as _
from ramstk.views.gtk3.widgets import RAMSTKTreePanel as RAMSTKTreePanel

class StakeholderTreePanel(RAMSTKTreePanel):
    _select_msg: str
    _tag: str
    _title: str
    _dic_row_loader: Dict[str, Callable]
    _on_edit_message: str
    dic_attribute_widget_map: Dict[str, List[Any]]
    def __init__(self) -> None: ...
    def do_load_affinity_groups(
        self, affinities: Dict[int, Tuple[str, str]]
    ) -> None: ...
    def do_load_stakeholders(self, stakeholders: Dict[int, str]) -> None: ...
    def _do_load_requirements(self, tree: treelib.Tree) -> None: ...
    def _on_insert(self, tree: treelib.Tree) -> None: ...
    def _on_module_switch(self, module: str = ...) -> None: ...
    _record_id: int
    _parent_id: int
    def _on_row_change(self, selection: Gtk.TreeSelection) -> None: ...
